""" ``ignite.contrib.metrics.regression.fractional_bias`` was moved to ``ignite.metrics.regression.fractional_bias``. # noqa
Note:
    ``ignite.contrib.metrics.regression.fractional_bias`` was moved to ``ignite.metrics.regression.fractional_bias``. # noqa
    Please refer to :mod:`~ignite.metrics.regression.fractional_bias`.
"""

import warnings

removed_in = "0.6.0"
deprecation_warning = (
    f"{__file__} has been moved to ignite/metrics/regression/fractional_bias.py"
    + f" and will be removed in version {removed_in}"
    + ".\n Please refer to the documentation for more details."
)
warnings.warn(deprecation_warning, DeprecationWarning, stacklevel=2)
from ignite.metrics.regression.fractional_bias import FractionalBias

__all__ = ["FractionalBias"]

FractionalBias = FractionalBias



# Vendored from py-rouge 1.1 (https://github.com/Diego999/py-rouge)
# Original license: Apache 2.0
import collections
import itertools
import os
import re
import urllib.request
from pathlib import Path

import nltk


class Rouge:
    DEFAULT_METRICS = {"rouge-n"}
    DEFAULT_N = 1
    STATS = ["f", "p", "r"]
    AVAILABLE_METRICS = {"rouge-n", "rouge-l", "rouge-w"}
    AVAILABLE_LENGTH_LIMIT_TYPES = {"words", "bytes"}
    REMOVE_CHAR_PATTERN = re.compile("[^A-Za-z0-9]")

    # Hack to not tokenize "cannot" to "can not" and consider them different as in the official ROUGE script
    KEEP_CANNOT_IN_ONE_WORD = re.compile("cannot")
    KEEP_CANNOT_IN_ONE_WORD_REVERSED = re.compile("_cannot_")

    WORDNET_KEY_VALUE = {}
    WORDNET_DB_FILEPATH = "wordnet_key_value.txt"
    WORDNET_DB_FILEPATH_SPECIAL_CASE = "wordnet_key_value_special_cases.txt"
    WORDNET_DB_DELIMITER = "|"
    STEMMER = None

    _WORDNET_BASE_URL = (
        "https://raw.githubusercontent.com/Diego999/py-rouge/c0176770f94c0b18cc9d116101f7859a04706208/rouge/"
    )
    WORDNET_DB_URL = _WORDNET_BASE_URL + "wordnet_key_value.txt"
    WORDNET_DB_SPECIAL_CASE_URL = _WORDNET_BASE_URL + "wordnet_key_value_special_cases.txt"

    def __init__(
        self,
        metrics=None,
        max_n=None,
        limit_length=True,
        length_limit=665,
        length_limit_type="bytes",
        apply_avg=True,
        apply_best=False,
        stemming=True,
        alpha=0.5,
        weight_factor=1.0,
        ensure_compatibility=True,
    ):
        self.metrics = metrics[:] if metrics is not None else Rouge.DEFAULT_METRICS
        for m in self.metrics:
            if m not in Rouge.AVAILABLE_METRICS:
                raise ValueError("Unknown metric '{}'".format(m))

        self.max_n = max_n if "rouge-n" in self.metrics else None
        if self.max_n is not None:
            index_rouge_n = self.metrics.index("rouge-n")
            del self.metrics[index_rouge_n]
            self.metrics += ["rouge-{}".format(n) for n in range(1, self.max_n + 1)]
        self.metrics = set(self.metrics)

        self.limit_length = limit_length
        if self.limit_length:
            if length_limit_type not in Rouge.AVAILABLE_LENGTH_LIMIT_TYPES:
                raise ValueError("Unknown length_limit_type '{}'".format(length_limit_type))

        self.length_limit = length_limit
        if self.length_limit == 0:
            self.limit_length = False
        self.length_limit_type = length_limit_type
        self.stemming = stemming

        self.apply_avg = apply_avg
        self.apply_best = apply_best
        self.alpha = alpha
        self.weight_factor = weight_factor
        if self.weight_factor <= 0:
            raise ValueError("ROUGE-W weight factor must greater than 0.")
        self.ensure_compatibility = ensure_compatibility

        # Only load stemming resources when stemming is actually enabled
        if self.stemming:
            if len(Rouge.WORDNET_KEY_VALUE) == 0:
                Rouge.load_wordnet_db(ensure_compatibility)
            if Rouge.STEMMER is None:
                Rouge.load_stemmer(ensure_compatibility)

    @staticmethod
    def load_stemmer(ensure_compatibility):
        Rouge.STEMMER = (
            nltk.stem.porter.PorterStemmer("ORIGINAL_ALGORITHM")
            if ensure_compatibility
            else nltk.stem.porter.PorterStemmer()
        )

    @staticmethod
    def load_wordnet_db(ensure_compatibility):
        files_to_load = [(Rouge.WORDNET_DB_FILEPATH, Rouge.WORDNET_DB_URL)]
        if ensure_compatibility:
            files_to_load.append((Rouge.WORDNET_DB_FILEPATH_SPECIAL_CASE, Rouge.WORDNET_DB_SPECIAL_CASE_URL))

        cache_dir = Path(os.path.expanduser("~/.cache/ignite/pyrouge"))
        cache_dir.mkdir(parents=True, exist_ok=True)

        for filename, url in files_to_load:
            cached_path = cache_dir / filename
            if not cached_path.exists():
                urllib.request.urlretrieve(url, str(cached_path))

            with open(cached_path, "r", encoding="utf-8") as fp:
                for line in fp:
                    k, v = line.strip().split(Rouge.WORDNET_DB_DELIMITER)
                    assert k not in Rouge.WORDNET_KEY_VALUE
                    Rouge.WORDNET_KEY_VALUE[k] = v

    @staticmethod
    def tokenize_text(text, language="english"):
        return nltk.word_tokenize(text, language)

    @staticmethod
    def split_into_sentences(text, ensure_compatibility, language="english"):
        if ensure_compatibility:
            return text.split("\n")
        else:
            return nltk.sent_tokenize(text, language)

    @staticmethod
    def stem_tokens(tokens):
        for i, token in enumerate(tokens):
            if len(token) > 0:
                if len(token) > 3:
                    if token in Rouge.WORDNET_KEY_VALUE:
                        token = Rouge.WORDNET_KEY_VALUE[token]
                    else:
                        token = Rouge.STEMMER.stem(token)
                    tokens[i] = token
        return tokens

    @staticmethod
    def _get_ngrams(n, text):
        ngram_set = collections.defaultdict(int)
        max_index_ngram_start = len(text) - n
        for i in range(max_index_ngram_start + 1):
            ngram_set[tuple(text[i : i + n])] += 1
        return ngram_set

    @staticmethod
    def _split_into_words(sentences):
        return list(itertools.chain(*[_.split() for _ in sentences]))

    @staticmethod
    def _get_word_ngrams_and_length(n, sentences):
        assert len(sentences) > 0
        assert n > 0
        tokens = Rouge._split_into_words(sentences)
        return Rouge._get_ngrams(n, tokens), tokens, len(tokens) - (n - 1)

    @staticmethod
    def _get_unigrams(sentences):
        assert len(sentences) > 0
        tokens = Rouge._split_into_words(sentences)
        unigram_set = collections.defaultdict(int)
        for token in tokens:
            unigram_set[token] += 1
        return unigram_set, len(tokens)

    @staticmethod
    def _compute_p_r_f_score(evaluated_count, reference_count, overlapping_count, alpha=0.5, weight_factor=1.0):
        precision = 0.0 if evaluated_count == 0 else overlapping_count / evaluated_count
        if weight_factor != 1.0:
            precision = precision ** (1.0 / weight_factor)
        recall = 0.0 if reference_count == 0 else overlapping_count / reference_count
        if weight_factor != 1.0:
            recall = recall ** (1.0 / weight_factor)
        f1_score = Rouge._compute_f_score(precision, recall, alpha)
        return {"f": f1_score, "p": precision, "r": recall}

    @staticmethod
    def _compute_f_score(precision, recall, alpha=0.5):
        return (
            0.0
            if (recall == 0.0 or precision == 0.0)
            else precision * recall / ((1 - alpha) * precision + alpha * recall)
        )

    @staticmethod
    def _compute_ngrams(evaluated_sentences, reference_sentences, n):
        if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
            raise ValueError("Collections must contain at least 1 sentence.")

        evaluated_ngrams, _, evaluated_count = Rouge._get_word_ngrams_and_length(n, evaluated_sentences)
        reference_ngrams, _, reference_count = Rouge._get_word_ngrams_and_length(n, reference_sentences)

        overlapping_ngrams = set(evaluated_ngrams.keys()).intersection(set(reference_ngrams.keys()))
        overlapping_count = 0
        for ngram in overlapping_ngrams:
            overlapping_count += min(evaluated_ngrams[ngram], reference_ngrams[ngram])

        return evaluated_count, reference_count, overlapping_count

    @staticmethod
    def _compute_ngrams_lcs(evaluated_sentences, reference_sentences, weight_factor=1.0):
        def _lcs(x, y):
            m = len(x)
            n = len(y)
            vals = collections.defaultdict(int)
            dirs = collections.defaultdict(int)

            for i in range(1, m + 1):
                for j in range(1, n + 1):
                    if x[i - 1] == y[j - 1]:
                        vals[i, j] = vals[i - 1, j - 1] + 1
                        dirs[i, j] = "|"
                    elif vals[i - 1, j] >= vals[i, j - 1]:
                        vals[i, j] = vals[i - 1, j]
                        dirs[i, j] = "^"
                    else:
                        vals[i, j] = vals[i, j - 1]
                        dirs[i, j] = "<"

            return vals, dirs

        def _wlcs(x, y, weight_factor):
            m = len(x)
            n = len(y)
            vals = collections.defaultdict(float)
            dirs = collections.defaultdict(int)
            lengths = collections.defaultdict(int)

            for i in range(1, m + 1):
                for j in range(1, n + 1):
                    if x[i - 1] == y[j - 1]:
                        length_tmp = lengths[i - 1, j - 1]
                        vals[i, j] = vals[i - 1, j - 1] + (length_tmp + 1) ** weight_factor - length_tmp**weight_factor
                        dirs[i, j] = "|"
                        lengths[i, j] = length_tmp + 1
                    elif vals[i - 1, j] >= vals[i, j - 1]:
                        vals[i, j] = vals[i - 1, j]
                        dirs[i, j] = "^"
                        lengths[i, j] = 0
                    else:
                        vals[i, j] = vals[i, j - 1]
                        dirs[i, j] = "<"
                        lengths[i, j] = 0

            return vals, dirs

        def _mark_lcs(mask, dirs, m, n):
            while m != 0 and n != 0:
                if dirs[m, n] == "|":
                    m -= 1
                    n -= 1
                    mask[m] = 1
                elif dirs[m, n] == "^":
                    m -= 1
                elif dirs[m, n] == "<":
                    n -= 1
                else:
                    raise UnboundLocalError("Illegal move")
            return mask

        if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
            raise ValueError("Collections must contain at least 1 sentence.")

        evaluated_unigrams_dict, evaluated_count = Rouge._get_unigrams(evaluated_sentences)
        reference_unigrams_dict, reference_count = Rouge._get_unigrams(reference_sentences)

        use_WLCS = weight_factor != 1.0
        if use_WLCS:
            evaluated_count = evaluated_count**weight_factor
            reference_count = 0

        overlapping_count = 0.0
        for reference_sentence in reference_sentences:
            reference_sentence_tokens = reference_sentence.split()
            if use_WLCS:
                reference_count += len(reference_sentence_tokens) ** weight_factor
            hit_mask = [0 for _ in range(len(reference_sentence_tokens))]

            for evaluated_sentence in evaluated_sentences:
                evaluated_sentence_tokens = evaluated_sentence.split()

                if use_WLCS:
                    _, lcs_dirs = _wlcs(reference_sentence_tokens, evaluated_sentence_tokens, weight_factor)
                else:
                    _, lcs_dirs = _lcs(reference_sentence_tokens, evaluated_sentence_tokens)
                _mark_lcs(hit_mask, lcs_dirs, len(reference_sentence_tokens), len(evaluated_sentence_tokens))

            overlapping_count_length = 0
            for ref_token_id, val in enumerate(hit_mask):
                if val == 1:
                    token = reference_sentence_tokens[ref_token_id]
                    if evaluated_unigrams_dict[token] > 0 and reference_unigrams_dict[token] > 0:
                        evaluated_unigrams_dict[token] -= 1
                        reference_unigrams_dict[ref_token_id] -= 1

                        if use_WLCS:
                            overlapping_count_length += 1
                            if (
                                ref_token_id + 1 < len(hit_mask) and hit_mask[ref_token_id + 1] == 0
                            ) or ref_token_id + 1 == len(hit_mask):
                                overlapping_count += overlapping_count_length**weight_factor
                                overlapping_count_length = 0
                        else:
                            overlapping_count += 1

        if use_WLCS:
            reference_count = reference_count**weight_factor

        return evaluated_count, reference_count, overlapping_count

    def get_scores(self, hypothesis, references):
        if isinstance(hypothesis, str):
            hypothesis, references = [hypothesis], [references]

        if type(hypothesis) != type(references):
            raise ValueError("'hyps' and 'refs' are not of the same type")

        if len(hypothesis) != len(references):
            raise ValueError("'hyps' and 'refs' do not have the same length")

        scores = {}
        has_rouge_n_metric = len([metric for metric in self.metrics if metric.split("-")[-1].isdigit()]) > 0
        if has_rouge_n_metric:
            scores = {**scores, **self._get_scores_rouge_n(hypothesis, references)}

        has_rouge_l_metric = len([metric for metric in self.metrics if metric.split("-")[-1].lower() == "l"]) > 0
        if has_rouge_l_metric:
            scores = {**scores, **self._get_scores_rouge_l_or_w(hypothesis, references, False)}

        has_rouge_w_metric = len([metric for metric in self.metrics if metric.split("-")[-1].lower() == "w"]) > 0
        if has_rouge_w_metric:
            scores = {**scores, **self._get_scores_rouge_l_or_w(hypothesis, references, True)}

        return scores

    def _get_scores_rouge_n(self, all_hypothesis, all_references):
        metrics = [metric for metric in self.metrics if metric.split("-")[-1].isdigit()]

        if self.apply_avg or self.apply_best:
            scores = {metric: {stat: 0.0 for stat in Rouge.STATS} for metric in metrics}
        else:
            scores = {
                metric: [{stat: [] for stat in Rouge.STATS} for _ in range(len(all_hypothesis))] for metric in metrics
            }

        for sample_id, (hypothesis, references) in enumerate(zip(all_hypothesis, all_references)):
            assert isinstance(hypothesis, str)
            has_multiple_references = False
            if isinstance(references, list):
                has_multiple_references = len(references) > 1
                if not has_multiple_references:
                    references = references[0]

            hypothesis = self._preprocess_summary_as_a_whole(hypothesis)
            references = (
                [self._preprocess_summary_as_a_whole(reference) for reference in references]
                if has_multiple_references
                else [self._preprocess_summary_as_a_whole(references)]
            )

            for metric in metrics:
                suffix = metric.split("-")[-1]
                n = int(suffix)

                if self.apply_avg:
                    total_hypothesis_ngrams_count = 0
                    total_reference_ngrams_count = 0
                    total_ngrams_overlapping_count = 0

                    for reference in references:
                        hypothesis_count, reference_count, overlapping_ngrams = Rouge._compute_ngrams(
                            hypothesis, reference, n
                        )
                        total_hypothesis_ngrams_count += hypothesis_count
                        total_reference_ngrams_count += reference_count
                        total_ngrams_overlapping_count += overlapping_ngrams

                    score = Rouge._compute_p_r_f_score(
                        total_hypothesis_ngrams_count,
                        total_reference_ngrams_count,
                        total_ngrams_overlapping_count,
                        self.alpha,
                    )

                    for stat in Rouge.STATS:
                        scores[metric][stat] += score[stat]
                else:
                    if self.apply_best:
                        best_current_score = None
                        for reference in references:
                            hypothesis_count, reference_count, overlapping_ngrams = Rouge._compute_ngrams(
                                hypothesis, reference, n
                            )
                            score = Rouge._compute_p_r_f_score(
                                hypothesis_count, reference_count, overlapping_ngrams, self.alpha
                            )
                            if best_current_score is None or score["r"] > best_current_score["r"]:
                                best_current_score = score

                        for stat in Rouge.STATS:
                            scores[metric][stat] += best_current_score[stat]
                    else:
                        for reference in references:
                            hypothesis_count, reference_count, overlapping_ngrams = Rouge._compute_ngrams(
                                hypothesis, reference, n
                            )
                            score = Rouge._compute_p_r_f_score(
                                hypothesis_count, reference_count, overlapping_ngrams, self.alpha
                            )
                            for stat in Rouge.STATS:
                                scores[metric][sample_id][stat].append(score[stat])

        if (self.apply_avg or self.apply_best) and len(all_hypothesis) > 1:
            for metric in metrics:
                for stat in Rouge.STATS:
                    scores[metric][stat] /= len(all_hypothesis)

        return scores

    def _get_scores_rouge_l_or_w(self, all_hypothesis, all_references, use_w=False):
        metric = "rouge-w" if use_w else "rouge-l"
        if self.apply_avg or self.apply_best:
            scores = {metric: {stat: 0.0 for stat in Rouge.STATS}}
        else:
            scores = {metric: [{stat: [] for stat in Rouge.STATS} for _ in range(len(all_hypothesis))]}

        for sample_id, (hypothesis_sentences, references_sentences) in enumerate(zip(all_hypothesis, all_references)):
            assert isinstance(hypothesis_sentences, str)
            has_multiple_references = False
            if isinstance(references_sentences, list):
                has_multiple_references = len(references_sentences) > 1
                if not has_multiple_references:
                    references_sentences = references_sentences[0]

            hypothesis_sentences = self._preprocess_summary_per_sentence(hypothesis_sentences)
            references_sentences = (
                [self._preprocess_summary_per_sentence(reference) for reference in references_sentences]
                if has_multiple_references
                else [self._preprocess_summary_per_sentence(references_sentences)]
            )

            if self.apply_avg:
                total_hypothesis_ngrams_count = 0
                total_reference_ngrams_count = 0
                total_ngrams_overlapping_count = 0

                for reference_sentences in references_sentences:
                    hypothesis_count, reference_count, overlapping_ngrams = Rouge._compute_ngrams_lcs(
                        hypothesis_sentences, reference_sentences, self.weight_factor if use_w else 1.0
                    )
                    total_hypothesis_ngrams_count += hypothesis_count
                    total_reference_ngrams_count += reference_count
                    total_ngrams_overlapping_count += overlapping_ngrams

                score = Rouge._compute_p_r_f_score(
                    total_hypothesis_ngrams_count,
                    total_reference_ngrams_count,
                    total_ngrams_overlapping_count,
                    self.alpha,
                    self.weight_factor,
                )

                for stat in Rouge.STATS:
                    scores[metric][stat] += score[stat]
            else:
                if self.apply_best:
                    best_current_score = None
                    best_current_score_wlcs = None
                    for reference_sentences in references_sentences:
                        hypothesis_count, reference_count, overlapping_ngrams = Rouge._compute_ngrams_lcs(
                            hypothesis_sentences, reference_sentences, self.weight_factor if use_w else 1.0
                        )
                        score = Rouge._compute_p_r_f_score(
                            hypothesis_count, reference_count, overlapping_ngrams, self.alpha, self.weight_factor
                        )

                        if use_w:
                            reference_count_for_score = reference_count ** (1.0 / self.weight_factor)
                            overlapping_ngrams_for_score = overlapping_ngrams
                            score_wlcs = (overlapping_ngrams_for_score / reference_count_for_score) ** (
                                1.0 / self.weight_factor
                            )

                            if best_current_score_wlcs is None or score_wlcs > best_current_score_wlcs:
                                best_current_score = score
                                best_current_score_wlcs = score_wlcs
                        else:
                            if best_current_score is None or score["r"] > best_current_score["r"]:
                                best_current_score = score

                    for stat in Rouge.STATS:
                        scores[metric][stat] += best_current_score[stat]
                else:
                    for reference_sentences in references_sentences:
                        hypothesis_count, reference_count, overlapping_ngrams = Rouge._compute_ngrams_lcs(
                            hypothesis_sentences, reference_sentences, self.weight_factor if use_w else 1.0
                        )
                        score = Rouge._compute_p_r_f_score(
                            hypothesis_count, reference_count, overlapping_ngrams, self.alpha, self.weight_factor
                        )

                        for stat in Rouge.STATS:
                            scores[metric][sample_id][stat].append(score[stat])

        if (self.apply_avg or self.apply_best) and len(all_hypothesis) > 1:
            for stat in Rouge.STATS:
                scores[metric][stat] /= len(all_hypothesis)

        return scores

    def _preprocess_summary_as_a_whole(self, summary):
        sentences = Rouge.split_into_sentences(summary, self.ensure_compatibility)

        if self.limit_length:
            if self.length_limit_type == "words":
                summary = " ".join(sentences)
                all_tokens = summary.split()
                summary = " ".join(all_tokens[: self.length_limit])
            elif self.length_limit_type == "bytes":
                summary = ""
                current_len = 0
                for sentence in sentences:
                    sentence = sentence.strip()
                    sentence_len = len(sentence)

                    if current_len + sentence_len < self.length_limit:
                        if current_len != 0:
                            summary += " "
                        summary += sentence
                        current_len += sentence_len
                    else:
                        if current_len > 0:
                            summary += " "
                        summary += sentence[: self.length_limit - current_len]
                        break
        else:
            summary = " ".join(sentences)

        summary = Rouge.REMOVE_CHAR_PATTERN.sub(" ", summary.lower()).strip()

        if self.ensure_compatibility:
            tokens = self.tokenize_text(Rouge.KEEP_CANNOT_IN_ONE_WORD.sub("_cannot_", summary))
        else:
            tokens = self.tokenize_text(Rouge.REMOVE_CHAR_PATTERN.sub(" ", summary))

        if self.stemming:
            self.stem_tokens(tokens)

        if self.ensure_compatibility:
            preprocessed_summary = [Rouge.KEEP_CANNOT_IN_ONE_WORD_REVERSED.sub("cannot", " ".join(tokens))]
        else:
            preprocessed_summary = [" ".join(tokens)]

        return preprocessed_summary

    def _preprocess_summary_per_sentence(self, summary):
        sentences = Rouge.split_into_sentences(summary, self.ensure_compatibility)

        if self.limit_length:
            final_sentences = []
            current_len = 0
            if self.length_limit_type == "words":
                for sentence in sentences:
                    tokens = sentence.strip().split()
                    tokens_len = len(tokens)
                    if current_len + tokens_len < self.length_limit:
                        sentence = " ".join(tokens)
                        final_sentences.append(sentence)
                        current_len += tokens_len
                    else:
                        sentence = " ".join(tokens[: self.length_limit - current_len])
                        final_sentences.append(sentence)
                        break
            elif self.length_limit_type == "bytes":
                for sentence in sentences:
                    sentence = sentence.strip()
                    sentence_len = len(sentence)
                    if current_len + sentence_len < self.length_limit:
                        final_sentences.append(sentence)
                        current_len += sentence_len
                    else:
                        sentence = sentence[: self.length_limit - current_len]
                        final_sentences.append(sentence)
                        break
            sentences = final_sentences

        final_sentences = []
        for sentence in sentences:
            sentence = Rouge.REMOVE_CHAR_PATTERN.sub(" ", sentence.lower()).strip()

            if self.ensure_compatibility:
                tokens = self.tokenize_text(Rouge.KEEP_CANNOT_IN_ONE_WORD.sub("_cannot_", sentence))
            else:
                tokens = self.tokenize_text(Rouge.REMOVE_CHAR_PATTERN.sub(" ", sentence))

            if self.stemming:
                self.stem_tokens(tokens)

            if self.ensure_compatibility:
                sentence = Rouge.KEEP_CANNOT_IN_ONE_WORD_REVERSED.sub("cannot", " ".join(tokens))
            else:
                sentence = " ".join(tokens)

            final_sentences.append(sentence)

        return final_sentences



import pytest

import torch
from scipy.stats import kendalltau

from ignite import distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.regression import KendallRankCorrelation


def test_zero_sample():
    with pytest.raises(
        NotComputableError, match="KendallRankCorrelation must have at least one example before it can be computed"
    ):
        metric = KendallRankCorrelation()
        metric.compute()


def test_wrong_y_pred_shape():
    with pytest.raises(ValueError, match=r"Input y_pred should have shape \(N,\) or \(N, 1\), but given"):
        metric = KendallRankCorrelation()
        y_pred = torch.arange(9).reshape(3, 3).float()
        y = torch.arange(3).unsqueeze(1).float()
        metric.update((y_pred, y))


def test_wrong_y_shape():
    with pytest.raises(ValueError, match=r"Input y should have shape \(N,\) or \(N, 1\), but given"):
        metric = KendallRankCorrelation()
        y_pred = torch.arange(3).unsqueeze(1).float()
        y = torch.arange(9).reshape(3, 3).float()
        metric.update((y_pred, y))


def test_wrong_y_pred_dtype():
    with pytest.raises(TypeError, match="Input y_pred dtype should be float 16, 32 or 64, but given"):
        metric = KendallRankCorrelation()
        y_pred = torch.arange(3).unsqueeze(1).long()
        y = torch.arange(3).unsqueeze(1).float()
        metric.update((y_pred, y))


def test_wrong_y_dtype():
    with pytest.raises(TypeError, match="Input y dtype should be float 16, 32 or 64, but given"):
        metric = KendallRankCorrelation()
        y_pred = torch.arange(3).unsqueeze(1).float()
        y = torch.arange(3).unsqueeze(1).long()
        metric.update((y_pred, y))


def test_wrong_variant():
    with pytest.raises(ValueError, match="variant accepts 'b' or 'c', got"):
        KendallRankCorrelation(variant="x")


@pytest.mark.parametrize("variant", ["b", "c"])
def test_kendall_correlation(variant: str, available_device):
    inputs = [torch.randn(4) for _ in range(4)]
    ground_truth = torch.randn(4)

    m = KendallRankCorrelation(variant=variant, device=available_device)
    assert m._device == torch.device(available_device)

    all_preds = []
    all_targets = []

    for pred in inputs:
        m.update((pred.to(device=available_device), ground_truth.to(device=available_device)))
        all_preds.append(pred)
        all_targets.append(ground_truth)

        concat_preds = torch.cat(all_preds).numpy()
        concat_targets = torch.cat(all_targets).numpy()

        expected = kendalltau(concat_preds, concat_targets, variant=variant).statistic
        assert m.compute() == pytest.approx(expected, rel=1e-4)


@pytest.fixture(params=list(range(2)))
def test_case(request):
    # correlated sample
    x = torch.randn(size=[50]).float()
    y = x + torch.randn_like(x) * 0.1

    return [
        (x, y, 1),
        (torch.rand(size=(50, 1)).float(), torch.rand(size=(50, 1)).float(), 10),
    ][request.param]


@pytest.mark.parametrize("n_times", range(5))
@pytest.mark.parametrize("variant", ["b", "c"])
@pytest.mark.parametrize(
    "test_case",
    [
        (torch.rand(size=(100,)), torch.rand(size=(100,)), 10),
        (torch.rand(size=(100, 1)), torch.rand(size=(100, 1)), 20),
    ],
)
def test_integration_kendall_rank_correlation(
    n_times: int, variant: str, test_case: tuple[torch.Tensor, torch.Tensor, int], available_device
):
    y_pred, y, batch_size = test_case

    def update_fn(engine: Engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        return y_pred[idx : idx + batch_size], y[idx : idx + batch_size]

    engine = Engine(update_fn)

    m = KendallRankCorrelation(variant=variant, device=available_device)
    assert m._device == torch.device(available_device)
    m.attach(engine, "kendall_tau")

    data = list(range(y_pred.shape[0] // batch_size))
    result = engine.run(data, max_epochs=1).metrics["kendall_tau"]

    expected = kendalltau(y_pred.cpu().view(-1).numpy(), y.cpu().view(-1).numpy(), variant=variant).statistic

    assert result == pytest.approx(expected, rel=2e-4)


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    @pytest.mark.parametrize("variant", ["b", "c"])
    def test_compute(self, variant: str):
        rank = idist.get_rank()
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        torch.manual_seed(10 + rank)
        for metric_device in metric_devices:
            m = KendallRankCorrelation(device=metric_device, variant=variant)

            y_pred = torch.rand(size=[100], device=device)
            y = torch.rand(size=[100], device=device)

            m.update((y_pred, y))

            y_pred = idist.all_gather(y_pred)
            y = idist.all_gather(y)

            np_y = y.cpu().numpy()
            np_y_pred = y_pred.cpu().numpy()

            np_ans = kendalltau(np_y_pred, np_y, variant=variant).statistic

            assert pytest.approx(np_ans, rel=2e-4) == m.compute()

    @pytest.mark.parametrize("n_epochs", [1, 2])
    @pytest.mark.parametrize("variant", ["b", "c"])
    def test_integration(self, n_epochs: int, variant: str):
        tol = 2e-4
        rank = idist.get_rank()
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        n_iters = 80
        batch_size = 16

        for metric_device in metric_devices:
            torch.manual_seed(12 + rank)

            y_true = torch.rand(size=(n_iters * batch_size,)).to(device)
            y_preds = torch.rand(size=(n_iters * batch_size,)).to(device)

            engine = Engine(
                lambda e, i: (
                    y_preds[i * batch_size : (i + 1) * batch_size],
                    y_true[i * batch_size : (i + 1) * batch_size],
                )
            )

            corr = KendallRankCorrelation(variant=variant, device=metric_device)
            corr.attach(engine, "kendall_tau")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_preds = idist.all_gather(y_preds)
            y_true = idist.all_gather(y_true)

            assert "kendall_tau" in engine.state.metrics

            res = engine.state.metrics["kendall_tau"]

            np_y = y_true.cpu().numpy()
            np_y_pred = y_preds.cpu().numpy()

            np_ans = kendalltau(np_y_pred, np_y, variant=variant).statistic

            assert pytest.approx(np_ans, rel=tol) == res



import torch
from collections.abc import Callable, Sequence

from ignite.exceptions import NotComputableError
from ignite.metrics.accuracy import _BaseClassification
from ignite.metrics.metric import sync_all_reduce
from ignite.metrics.fairness.base import SubgroupDifference

__all__ = ["DemographicParityDifference", "SelectionRate"]


class SelectionRate(_BaseClassification):
    """Calculates the selection rate (rate of positive predictions).

    - ``update`` must receive output of the form ``(y_pred, y)``.
    - `y_pred` must be in the following shape (batch_size, num_categories, ...) or (batch_size, ...).
    - `y` must be in the following shape (batch_size, ...).

    Args:
        output_transform: a callable that is used to transform the engine output.
        is_multilabel: if True, multilabel selection rate is calculated. By default, False.
        device: specifies the computation device.
        skip_unrolling: specifies whether output should be unrolled before being fed to update method.

    Examples:

        .. include:: defaults.rst
            :start-after: :orphan:

        .. testcode::

            metric = SelectionRate()
            metric.attach(default_evaluator, 'selection_rate')
            y_pred = torch.tensor([[0.1, 0.9], [0.2, 0.8], [0.9, 0.1], [0.9, 0.1]])
            y_true = torch.tensor([1, 1, 0, 0])  # ignored
            state = default_evaluator.run([[y_pred, y_true]])
            print(state.metrics['selection_rate'])

        .. testoutput::

            tensor([0.5000, 0.5000])

    .. versionadded:: 0.6.0
    """

    _state_dict_all_req_keys = ("_num_positives", "_num_examples")

    def __init__(
        self,
        output_transform: Callable = lambda x: x,
        is_multilabel: bool = False,
        device: str | torch.device = torch.device("cpu"),
        skip_unrolling: bool = False,
    ):
        self._num_positives: torch.Tensor = torch.tensor(0.0, device=device)
        super().__init__(
            output_transform=output_transform, is_multilabel=is_multilabel, device=device, skip_unrolling=skip_unrolling
        )

    def reset(self) -> None:
        self._num_positives = torch.tensor(0.0, device=self._device)
        self._num_examples = 0
        super().reset()

    def update(self, output: Sequence[torch.Tensor]) -> None:
        self._check_shape(output)
        self._check_type(output)
        y_pred = output[0].detach()

        if self._type == "binary":
            positives = torch.bincount(y_pred.view(-1).to(torch.long), minlength=2).float()
            total = y_pred.numel()
        elif self._type == "multiclass":
            if self._num_classes is None:
                raise RuntimeError("num_classes must be set for multiclass data.")
            predicted_classes = torch.argmax(y_pred, dim=1)
            positives = torch.bincount(predicted_classes.view(-1), minlength=self._num_classes).float()
            total = predicted_classes.numel()
        elif self._type == "multilabel":
            if self._num_classes is None:
                raise RuntimeError("num_classes must be set for multilabel data.")
            num_classes: int = self._num_classes
            positives = y_pred.movedim(1, -1).reshape(-1, num_classes).sum(dim=0).float()
            total = int(y_pred.numel() / num_classes)
        else:
            raise ValueError(f"Unexpected type: {self._type}")

        if self._num_positives.ndim == 0:
            self._num_positives = positives.to(self._device)
        else:
            self._num_positives = self._num_positives + positives.to(self._device)
        self._num_examples += total

    @sync_all_reduce("_num_examples", "_num_positives")
    def compute(self) -> torch.Tensor:
        """Computes the selection rate.

        Returns:
            The selection rate for each category/label.
        """
        if self._num_examples == 0 or self._num_positives.ndim == 0:
            raise NotComputableError("SelectionRate must have at least one example before it can be computed.")
        return self._num_positives / self._num_examples


class DemographicParityDifference(SubgroupDifference):
    r"""Calculates the Demographic Parity Difference.

    This metric computes the selection rate (the rate of positive predictions) for each unique
    subgroup in the dataset and returns the maximum difference in selection rates between any
    two subgroups.

    A lower value indicates that the model predicts the positive outcome at roughly the same rate
    across all subgroups, a standard definition of fairness. This metric is referred to as
    *Group Fairness / Statistical Parity* in the fairness literature.

    - ``update`` must receive output of the form ``(y_pred, y, group_labels)`` or
      ``{'y_pred': y_pred, 'y': y, 'group_labels': group_labels}``.
    - `y_pred` must be in the following shape (batch_size, num_categories, ...) or (batch_size, ...).
    - `y` must be in the following shape (batch_size, ...).
    - `group_labels` must be a 1D tensor of shape (batch_size,) containing discrete labels.

    Args:
        groups: a sequence of unique group identifiers.
        is_multilabel: if True, multilabel selection rate is calculated. By default, False.
        output_transform: a callable that is used to transform the
            :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
            form expected by the metric.
        device: specifies which device updates are accumulated on. Setting the
            metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
            non-blocking. By default, CPU.

    Examples:
        To use with ``Engine`` and ``process_function``, simply attach the metric instance to the engine.
        The output of the engine's ``process_function`` needs to be in the format of
        ``(y_pred, y, group_labels)``.

        .. include:: defaults.rst
            :start-after: :orphan:

        .. testcode::

            metric = DemographicParityDifference(groups=[0, 1])
            metric.attach(default_evaluator, 'demographic_parity_diff')

            # Predictions for 4 items:
            # Items 1 and 2 are predicted as class 1 (index 1 has highest prob)
            # Items 3 and 4 are predicted as class 0 (index 0 has highest prob)
            y_pred = torch.tensor([[0.1, 0.9], [0.2, 0.8], [0.9, 0.1], [0.9, 0.1]])

            # Targets (Not actually used for parity, but required by API)
            y_true = torch.tensor([1, 1, 0, 0])

            # Subgroups
            group_labels = torch.tensor([0, 0, 1, 1])

            # Subgroup 0: 2 positive predictions / 2 total = 1.0 selection rate
            # Subgroup 1: 0 positive predictions / 2 total = 0.0 selection rate

            state = default_evaluator.run([[y_pred, y_true, group_labels]])
            print(state.metrics['demographic_parity_diff'])

        .. testoutput::

            1.0

    .. versionadded:: 0.6.0

    References:
        - Verma & Rubin, `Fairness Definitions Explained
          <https://fairware.cs.umass.edu/papers/Verma.pdf>`_, 2018.
    """

    def __init__(
        self,
        groups: Sequence[int],
        is_multilabel: bool = False,
        output_transform: Callable = lambda x: x,
        device: torch.device | str = torch.device("cpu"),
    ) -> None:
        sr = SelectionRate(is_multilabel=is_multilabel, device=device)
        super().__init__(base_metric=sr, groups=groups, output_transform=output_transform, device=device)



import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as F

import ignite.distributed as idist


class Net(nn.Module):
    def __init__(self, l1=120, l2=84):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, l1)
        self.fc2 = nn.Linear(l1, l2)
        self.fc3 = nn.Linear(l2, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


def load_data(data_dir="./data"):
    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(
                (0.4914, 0.4822, 0.4465),
                (0.2023, 0.1994, 0.2010),
            ),
        ]
    )

    train_set = torchvision.datasets.CIFAR10(root=data_dir, train=True, download=True, transform=transform)
    test_set = torchvision.datasets.CIFAR10(root=data_dir, train=False, download=True, transform=transform)

    return train_set, test_set


def get_data_loaders(batch_size, data_dir="./data", num_workers=8):
    train_set, _ = load_data(data_dir)

    train_size = int(len(train_set) * 0.8)
    val_size = len(train_set) - train_size

    train_subset, val_subset = torch.utils.data.random_split(train_set, [train_size, val_size])

    train_loader = idist.auto_dataloader(
        train_subset, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True
    )
    val_loader = idist.auto_dataloader(val_subset, batch_size=batch_size, shuffle=False, num_workers=num_workers)

    return train_loader, val_loader


def get_test_loader(batch_size, data_dir="./data", num_workers=8):
    _, test_set = load_data(data_dir)

    test_loader = idist.auto_dataloader(test_set, batch_size=batch_size, shuffle=False, num_workers=num_workers)

    return test_loader



import pytest
import torch
from torch.testing import assert_close

import ignite.distributed as idist
from fairlearn.metrics import demographic_parity_difference
from ignite.exceptions import NotComputableError
from ignite.metrics.fairness.demographic_parity import DemographicParityDifference, SelectionRate


def test_demographic_parity_difference_empty() -> None:
    """Tests if NotComputableError is raised when no data is provided."""
    metric = DemographicParityDifference(groups=[0, 1])
    with pytest.raises(NotComputableError, match="Fairness metrics must have at least one example"):
        metric.compute()


def test_demographic_parity_difference_single_group() -> None:
    """Tests if NotComputableError is raised when only one subgroup is present."""
    metric = DemographicParityDifference(groups=[0])
    y_pred = torch.tensor([[0.9, 0.1], [0.8, 0.2]])
    y = torch.tensor([0, 0])
    group_labels = torch.tensor([0, 0])
    metric.update((y_pred, y, group_labels))
    with pytest.raises(NotComputableError, match="Fairness metrics require at least two unique subgroups"):
        metric.compute()


def test_demographic_parity_difference_binary_probs_shape_B(available_device) -> None:
    """Tests DemographicParityDifference with shape (B,) thresholded inputs."""
    metric = DemographicParityDifference(groups=[0, 1], device=available_device)
    # y_pred is (B,) already thresholded
    # Group 0: 1 pos / 2 total = 0.5
    # Group 1: 0 pos / 2 total = 0.0
    y_pred = torch.tensor([1, 0, 0, 0], device=available_device)
    y = torch.tensor([0, 0, 0, 0], device=available_device)  # ignored
    groups = torch.tensor([0, 0, 1, 1], device=available_device)

    metric.update((y_pred, y, groups))
    assert_close(metric.compute(), 0.5)


def test_demographic_parity_difference_binary_probs_shape_B_1(available_device) -> None:
    """Tests DemographicParityDifference with shape (B, 1) thresholded inputs."""
    metric = DemographicParityDifference(groups=[0, 1], device=available_device)
    # y_pred is (B, 1) already thresholded
    y_pred = torch.tensor([[1], [0], [1], [0]], device=available_device)
    y = torch.tensor([0, 0, 0, 0], device=available_device)
    groups = torch.tensor([0, 0, 1, 1], device=available_device)
    metric.update((y_pred, y, groups))
    # G0 selection rate: 0.5, G1 selection rate: 0.5 -> Diff 0.0
    assert_close(metric.compute(), 0.0)


def test_demographic_parity_difference_multiclass(available_device) -> None:
    """Tests DemographicParityDifference with multiclass logits."""
    metric = DemographicParityDifference(groups=[0, 1], device=available_device)
    # y_pred is (B, C)
    # G0 selection rates: [0.5, 0.5, 0.0]
    # G1 selection rates: [0.5, 0.0, 0.5]
    y_pred = torch.tensor(
        [
            [0.8, 0.1, 0.1],  # pred class 0
            [0.1, 0.8, 0.1],  # pred class 1
            [0.8, 0.1, 0.1],  # pred class 0
            [0.1, 0.1, 0.8],  # pred class 2
        ],
        device=available_device,
    )
    y = torch.tensor([0, 0, 0, 0], device=available_device)
    groups = torch.tensor([0, 0, 1, 1], device=available_device)

    metric.update((y_pred, y, groups))
    # Disparities: Class 0: 0.0, Class 1: 0.5, Class 2: 0.5
    assert_close(metric.compute(), 0.5)


def test_demographic_parity_difference_multilabel(available_device) -> None:
    """Tests DemographicParityDifference with multilabel data."""
    metric = DemographicParityDifference(groups=[0, 1], is_multilabel=True, device=available_device)
    # y_pred is (B, C) indicators
    # G0: [1, 1, 0], [0, 0, 0] -> rates: [0.5, 0.5, 0.0]
    # G1: [1, 1, 1], [0, 1, 0] -> rates: [0.5, 1.0, 0.5]
    y_pred = torch.tensor([[1, 1, 0], [0, 0, 0], [1, 1, 1], [0, 1, 0]], device=available_device)
    y = torch.tensor([[0, 0, 0]] * 4, device=available_device)
    groups = torch.tensor([0, 0, 1, 1], device=available_device)

    metric.update((y_pred, y, groups))
    # Disparities: C0: 0.0, C1: 0.5, C2: 0.5
    assert_close(metric.compute(), 0.5)


def test_compare_demographic_parity_with_fairlearn(available_device) -> None:
    """Verifies DemographicParityDifference matches Fairlearn's demographic_parity_difference"""
    ignite_metric = DemographicParityDifference(groups=[0, 1], device=available_device)

    # Multi-class case
    # G0: [0, 1, 0], G1: [1, 1, 1]
    y_pred_probs = torch.tensor(
        [
            [0.9, 0.1, 0.0],
            [0.1, 0.8, 0.1],
            [0.9, 0.1, 0.0],
            [0.1, 0.8, 0.1],
            [0.1, 0.8, 0.1],
            [0.1, 0.8, 0.1],
        ],
        device=available_device,
    )
    y_true = torch.zeros(6, device=available_device)  # ignored by SelectionRate
    group_labels = torch.tensor([0, 0, 0, 1, 1, 1], device=available_device)

    ignite_metric.update((y_pred_probs, y_true, group_labels))
    ignite_result = ignite_metric.compute()

    # Fairlearn verification for multiclass
    # We calculate demographic parity difference for each class independently (one-vs-rest)
    # and take the max, which is how Ignite's SubgroupDifference handles vector outputs.
    y_pred_classes = torch.argmax(y_pred_probs, dim=1).cpu().numpy()
    fairlearn_max_diff = 0.0
    for c in range(3):
        y_pred_bin = (y_pred_classes == c).astype(int)
        diff = demographic_parity_difference(
            y_true=y_true.cpu().numpy(), y_pred=y_pred_bin, sensitive_features=group_labels.cpu().numpy()
        )
        fairlearn_max_diff = max(fairlearn_max_diff, diff)

    assert_close(ignite_result, float(fairlearn_max_diff))

    # Simple binary case
    y_pred_binary = torch.tensor([1, 0, 1, 0, 0, 0], device=available_device)
    group_labels_binary = torch.tensor([0, 0, 0, 1, 1, 1], device=available_device)

    ignite_metric.reset()
    ignite_metric.update((y_pred_binary, y_true, group_labels_binary))
    ignite_res_bin = ignite_metric.compute()

    fairlearn_res_bin = demographic_parity_difference(
        y_true=y_true.cpu().numpy(),
        y_pred=y_pred_binary.cpu().numpy(),
        sensitive_features=group_labels_binary.cpu().numpy(),
    )

    assert_close(ignite_res_bin, float(fairlearn_res_bin))


def test_selection_rate_binary(available_device) -> None:
    """Tests SelectionRate for basic binary predictions."""
    metric = SelectionRate(device=available_device)
    # 2 positives out of 4
    y_pred = torch.tensor([1, 0, 1, 0], device=available_device)
    y = torch.tensor([0, 0, 0, 0], device=available_device)
    metric.update((y_pred, y))
    res = metric.compute()
    assert_close(res, torch.tensor([0.5, 0.5], device=available_device))


def test_selection_rate_multiclass(available_device) -> None:
    """Tests SelectionRate for multiclass predictions."""
    metric = SelectionRate(device=available_device)
    # 0: 2 picks, 1: 1 pick, 2: 1 pick. Total 4.
    y_pred = torch.tensor([[0.8, 0.1, 0.1], [0.8, 0.1, 0.1], [0.1, 0.8, 0.1], [0.1, 0.1, 0.8]], device=available_device)
    y = torch.tensor([0, 0, 0, 0], device=available_device)
    metric.update((y_pred, y))
    res = metric.compute()
    # proportions: [0.5, 0.25, 0.25]
    assert_close(res, torch.tensor([0.5, 0.25, 0.25], device=available_device))


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_compute_demographic_parity_difference(self) -> None:
        rank = idist.get_rank()
        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            metric = DemographicParityDifference(groups=[0, 1], device=metric_device)

            y = torch.tensor([0, 0], device=device)
            groups = torch.tensor([0, 1], device=device)

            if rank == 0:
                # G0: pos, G1: neg
                y_pred = torch.tensor([[0.2, 0.8], [0.8, 0.2]], device=device)
            else:
                # G0: pos, G1: neg
                y_pred = torch.tensor([[0.1, 0.9], [0.9, 0.1]], device=device)

            metric.update((y_pred, y, groups))
            res = metric.compute()

            # Across all ranks:
            # Group 0 selected rate: 1.0
            # Group 1 selected rate: 0.0
            # Max Disparity = 1.0
            assert_close(res, 1.0)


def test_demographic_parity_difference_validation() -> None:
    """Tests input validation for DemographicParityDifference."""
    metric = DemographicParityDifference(groups=[0, 1])

    # Valid multiclass logits
    y_pred = torch.tensor(
        [
            [0.1, 0.9, 0.0],
            [0.1, 0.9, 0.0],  # Group 0: all class 1
            [0.9, 0.1, 0.0],
            [0.9, 0.1, 0.0],  # Group 1: all class 0
        ]
    )
    y = torch.tensor([1, 1, 0, 0])
    groups = torch.tensor([0, 0, 1, 1])
    metric.update((y_pred, y, groups))
    # Class 0: Group 0 rate = 0, Group 1 rate = 1.0 -> Diff = 1.0
    # Class 1: Group 0 rate = 1.0, Group 1 rate = 0 -> Diff = 1.0
    assert metric.compute() == 1.0

    metric.reset()

    # Valid binary classes (0/1)
    y_pred = torch.tensor([1, 1, 0, 0])
    y = torch.tensor([1, 1, 0, 0])
    groups = torch.tensor([0, 0, 1, 1])
    metric.update((y_pred, y, groups))
    assert metric.compute() == 1.0

    metric.reset()

    # Invalid binary probabilities
    y_pred = torch.tensor([0.6, 0.4, 0.8, 0.1])
    y = torch.tensor([1, 0, 1, 0])
    groups = torch.tensor([0, 0, 1, 1])
    with pytest.raises(ValueError, match="y_pred must be comprised of 0's and 1's"):
        metric.update((y_pred, y, groups))


def test_demographic_parity_multilabel_validation() -> None:
    """Tests input validation for DemographicParityDifference multilabel data."""
    metric = DemographicParityDifference(groups=[0, 1], is_multilabel=True)

    # Valid multilabel (0/1)
    # y_pred is (B, C) indicators
    y_pred = torch.tensor([[1, 0], [1, 1], [0, 0], [0, 1]])
    y = torch.tensor([[0, 0], [0, 0], [0, 0], [0, 0]])  # Ignored by SelectionRate
    groups = torch.tensor([0, 0, 1, 1])
    metric.update((y_pred, y, groups))
    assert metric.compute() >= 0

    metric.reset()

    # Invalid multilabel (not 0/1)
    y_pred = torch.tensor([[0.6, 0.4], [0.8, 0.1]])
    y = torch.tensor([[1, 0], [1, 1]])
    groups = torch.tensor([0, 1])
    with pytest.raises(ValueError, match="y_pred must be comprised of 0's and 1's"):
        metric.update((y_pred, y, groups))


def test_demographic_parity_shape_mismatch_validation() -> None:
    """Tests validation for shape mismatches between y and y_pred."""
    metric = DemographicParityDifference(groups=[0, 1])

    # y is (B,), y_pred is (B, C) - OK
    metric.update((torch.randn(4, 3), torch.randint(0, 3, (4,)), torch.zeros(4)))

    # y is (B,), y_pred is (B, C, H, W) - Error (dimension mismatch)
    with pytest.raises(ValueError, match="y must have shape of"):
        metric.update((torch.randn(4, 3, 2, 2), torch.randint(0, 3, (4,)), torch.zeros(4)))


def test_demographic_parity_batch_size_mismatch_validation() -> None:
    """Tests validation for batch size mismatches across y_pred, y, and groups."""
    metric = DemographicParityDifference(groups=[0, 1])

    # y_pred (4), y (3) - Batch size mismatch
    with pytest.raises(ValueError, match="y_pred, y, and group_labels must have the same batch size"):
        metric.update((torch.randn(4, 3), torch.randint(0, 3, (3,)), torch.zeros(4)))

    # groups (3), y_pred (4) - Batch size mismatch
    with pytest.raises(ValueError, match="y_pred, y, and group_labels must have the same batch size"):
        metric.update((torch.randn(4, 3), torch.randint(0, 3, (4,)), torch.zeros(3)))


def test_demographic_parity_type_switch_validation() -> None:
    """Tests if RuntimeError is raised when input type changes mid-epoch."""
    metric = DemographicParityDifference(groups=[0, 1])

    # First batch: binary
    metric.update((torch.tensor([1, 0]), torch.tensor([1, 0]), torch.tensor([0, 1])))

    # Second batch: multiclass - Should raise RuntimeError
    with pytest.raises(RuntimeError, match="Input data type has changed from binary to multiclass"):
        metric.update((torch.randn(2, 3), torch.tensor([0, 1]), torch.tensor([0, 1])))

    # Third batch: multiclass (3) then multiclass (5) - Should raise ValueError
    metric.reset()
    metric.update((torch.randn(4, 3), torch.randint(0, 3, (4,)), torch.zeros(4)))
    with pytest.raises(ValueError, match="Input data number of classes has changed"):
        metric.update((torch.randn(4, 5), torch.randint(0, 5, (4,)), torch.zeros(4)))


def test_demographic_parity_difference_checkpointing() -> None:
    """Verifies that the state of SubgroupDifference is correctly preserved across state_dict calls."""
    metric = DemographicParityDifference(groups=[0, 1])

    # 1. Update with some data for Group 0
    y_pred_0 = torch.tensor([[0.1, 0.9], [0.1, 0.9]])
    y_true_0 = torch.tensor([1, 1])
    groups_0 = torch.tensor([0, 0])
    metric.update((y_pred_0, y_true_0, groups_0))

    # 2. Save the state
    state = metric.state_dict()

    # 3. Create a NEW metric instance and load the state
    new_metric = DemographicParityDifference(groups=[0, 1])
    new_metric.load_state_dict(state)

    # 4. Update the NEW metric with data for Group 1 only
    y_pred_1 = torch.tensor([[0.9, 0.1], [0.9, 0.1]])
    y_true_1 = torch.tensor([1, 1])
    groups_1 = torch.tensor([1, 1])
    new_metric.update((y_pred_1, y_true_1, groups_1))

    # 5. Compute the result on the new metric
    # The new metric should have 1.0 selection rate for Group 0 (from state)
    # and 0.0 selection rate for Group 1 (from new update).
    # Result: 1.0 - 0.0 = 1.0
    assert new_metric.compute() == 1.0



from unittest.mock import patch

import pytest
import sklearn
import torch
from sklearn.metrics import matthews_corrcoef

from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics import MatthewsCorrCoef

torch.manual_seed(12)


@pytest.fixture()
def mock_no_sklearn():
    with patch.dict("sys.modules", {"sklearn.metrics": None}):
        yield sklearn


def test_no_sklearn(mock_no_sklearn):
    with pytest.raises(ModuleNotFoundError, match=r"This metric module requires scikit-learn to be installed."):
        MatthewsCorrCoef()


def test_no_update():
    mcc = MatthewsCorrCoef()

    with pytest.raises(
        NotComputableError, match=r"EpochMetric must have at least one example before it can be computed"
    ):
        mcc.compute()


def test_input_types():
    mcc = MatthewsCorrCoef()
    mcc.reset()
    output1 = (torch.rand(4, 3), torch.randint(0, 2, size=(4, 3), dtype=torch.long))
    mcc.update(output1)

    with pytest.raises(ValueError, match=r"Incoherent types between input y_pred and stored predictions"):
        mcc.update((torch.randint(0, 5, size=(4, 3)), torch.randint(0, 2, size=(4, 3))))

    with pytest.raises(ValueError, match=r"Incoherent types between input y and stored targets"):
        mcc.update((torch.rand(4, 3), torch.randint(0, 2, size=(4, 3)).to(torch.int32)))

    with pytest.raises(ValueError, match=r"Incoherent types between input y_pred and stored predictions"):
        mcc.update((torch.randint(0, 2, size=(10,)).long(), torch.randint(0, 2, size=(10, 5)).long()))


def test_check_shape():
    mcc = MatthewsCorrCoef()

    with pytest.raises(ValueError, match=r"Predictions should be of shape"):
        mcc._check_shape((torch.tensor(0), torch.tensor(0)))

    with pytest.raises(ValueError, match=r"Predictions should be of shape"):
        mcc._check_shape((torch.rand(4, 3, 1), torch.rand(4, 3)))

    with pytest.raises(ValueError, match=r"Targets should be of shape"):
        mcc._check_shape((torch.rand(4, 3), torch.rand(4, 3, 1)))


@pytest.fixture(params=range(4))
def test_data_binary(request):
    return [
        # Binary input data of shape (N,) or (N, 1)
        (torch.randint(0, 2, size=(10,)).long(), torch.randint(0, 2, size=(10,)).long(), 1),
        (torch.randint(0, 2, size=(10, 1)).long(), torch.randint(0, 2, size=(10, 1)).long(), 1),
        # updated batches
        (torch.randint(0, 2, size=(50,)).long(), torch.randint(0, 2, size=(50,)).long(), 16),
        (torch.randint(0, 2, size=(50, 1)).long(), torch.randint(0, 2, size=(50, 1)).long(), 16),
    ][request.param]


@pytest.mark.parametrize("n_times", range(2))
def test_binary_input(n_times, test_data_binary, available_device):
    y_pred, y, batch_size = test_data_binary
    mcc = MatthewsCorrCoef(device=available_device)
    assert mcc._device == torch.device(available_device)

    mcc.reset()
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            mcc.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        mcc.update((y_pred, y))

    np_y = y.numpy().ravel()
    np_y_pred = y_pred.numpy().ravel()

    assert isinstance(mcc.compute(), float)
    assert matthews_corrcoef(np_y, np_y_pred) == pytest.approx(mcc.compute())


@pytest.fixture(params=range(2))
def test_data_multiclass(request):
    return [
        # Multiclass input data of shape (N,)
        (torch.randint(0, 5, size=(10,)).long(), torch.randint(0, 5, size=(10,)).long(), 1),
        # updated batches
        (torch.randint(0, 5, size=(50,)).long(), torch.randint(0, 5, size=(50,)).long(), 16),
    ][request.param]


@pytest.mark.parametrize("n_times", range(2))
def test_multiclass_input(n_times, test_data_multiclass, available_device):
    y_pred, y, batch_size = test_data_multiclass
    mcc = MatthewsCorrCoef(device=available_device)
    assert mcc._device == torch.device(available_device)

    mcc.reset()
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            mcc.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        mcc.update((y_pred, y))

    np_y = y.numpy().ravel()
    np_y_pred = y_pred.numpy().ravel()

    assert isinstance(mcc.compute(), float)
    assert matthews_corrcoef(np_y, np_y_pred) == pytest.approx(mcc.compute())


def test_integration(available_device):
    y_pred = torch.tensor([1, 0, 1, 1])
    y_true = torch.tensor([1, 1, 0, 1])

    def update_fn(engine, batch):
        return y_pred, y_true

    evaluator = Engine(update_fn)
    mcc = MatthewsCorrCoef(device=available_device)
    mcc.attach(evaluator, "mcc")

    state = evaluator.run([None], max_epochs=1)

    assert state.metrics["mcc"] == pytest.approx(matthews_corrcoef(y_true.numpy(), y_pred.numpy()))



import numpy as np
import pytest
import torch
from torch import Tensor

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics import CosineSimilarity


def test_zero_sample():
    cos_sim = CosineSimilarity()
    with pytest.raises(
        NotComputableError, match=r"CosineSimilarity must have at least one example before it can be computed"
    ):
        cos_sim.compute()


@pytest.fixture(params=list(range(4)))
def test_case(request):
    torch.manual_seed(0)  # For reproducibility

    eps = float(torch.empty(1).uniform_(-8, 0).exp())  # 10 ** uniform(-8, 0)

    return [
        (torch.randn((100, 50)), torch.randn((100, 50)), eps, 1),
        (torch.normal(1.0, 2.0, size=(100, 10)), torch.normal(3.0, 4.0, size=(100, 10)), eps, 1),
        (torch.rand((100, 128)), torch.rand((100, 128)), eps, 16),
        (torch.normal(0.0, 5.0, size=(100, 30)), torch.normal(5.0, 1.0, size=(100, 30)), eps, 16),
    ][request.param]


@pytest.mark.parametrize("n_times", range(5))
def test_compute(n_times, test_case: tuple[Tensor, Tensor, float, int], available_device):
    y_pred, y, eps, batch_size = test_case

    cos = CosineSimilarity(eps=eps, device=available_device)
    assert cos._device == torch.device(available_device)

    cos.reset()
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            cos.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        cos.update((y_pred, y))

    y_norm = torch.clamp(torch.norm(y, dim=1, keepdim=True), min=eps)
    y_pred_norm = torch.clamp(torch.norm(y_pred, dim=1, keepdim=True), min=eps)

    cosine_sim = torch.sum((y / y_norm) * (y_pred / y_pred_norm), dim=1)
    expected = cosine_sim.mean().item()

    result = cos.compute()

    assert isinstance(result, float)
    assert pytest.approx(expected, rel=2e-5) == result


def test_accumulator_detached(available_device):
    cos = CosineSimilarity(device=available_device)
    assert cos._device == torch.device(available_device)

    y_pred = torch.tensor([[2.0, 3.0], [-2.0, 1.0]], dtype=torch.float)
    y = torch.ones(2, 2, dtype=torch.float)
    cos.update((y_pred, y))

    assert not cos._sum_of_cos_similarities.requires_grad


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_integration(self):
        tol = 2e-5
        n_iters = 100
        batch_size = 10
        n_dims = 100

        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            y_true = torch.randn((n_iters * batch_size, n_dims)).float().to(device)
            y_preds = torch.normal(2.0, 3.0, size=(n_iters * batch_size, n_dims)).float().to(device)

            engine = Engine(
                lambda e, i: (
                    y_preds[i * batch_size : (i + 1) * batch_size],
                    y_true[i * batch_size : (i + 1) * batch_size],
                )
            )

            m = CosineSimilarity(device=metric_device)
            m.attach(engine, "cosine_similarity")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=1)

            y_preds = idist.all_gather(y_preds)
            y_true = idist.all_gather(y_true)

            assert "cosine_similarity" in engine.state.metrics
            res = engine.state.metrics["cosine_similarity"]

            y_true_np = y_true.cpu().numpy()
            y_preds_np = y_preds.cpu().numpy()
            y_true_norm = np.clip(np.linalg.norm(y_true_np, axis=1, keepdims=True), 1e-8, None)
            y_preds_norm = np.clip(np.linalg.norm(y_preds_np, axis=1, keepdims=True), 1e-8, None)
            true_res = np.sum((y_true_np / y_true_norm) * (y_preds_np / y_preds_norm), axis=1)
            true_res = np.mean(true_res)

            assert pytest.approx(res, rel=tol) == true_res

    def test_accumulator_device(self):
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)
        for metric_device in metric_devices:
            cos = CosineSimilarity(device=metric_device)

            for dev in (cos._device, cos._sum_of_cos_similarities.device):
                assert dev == metric_device, f"{type(dev)}:{dev} vs {type(metric_device)}:{metric_device}"

            y_pred = torch.tensor([[2.0, 3.0], [-2.0, 1.0]]).float()
            y = torch.ones(2, 2).float()
            cos.update((y_pred, y))

            for dev in (cos._device, cos._sum_of_cos_similarities.device):
                assert dev == metric_device, f"{type(dev)}:{dev} vs {type(metric_device)}:{metric_device}"



import os
from importlib.util import find_spec
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
import torch
from packaging.version import Version
from pytest import approx
from torch.nn.functional import mse_loss
from torch.optim import SGD

# Min pytorch version support is 2.2.2
try:
    from torch.amp import GradScaler
except ImportError:
    from torch.cuda.amp import GradScaler


import ignite.distributed as idist
from ignite.engine import (
    Engine,
    Events,
    _check_arg,
    create_supervised_evaluator,
    create_supervised_trainer,
    supervised_evaluation_step,
    supervised_evaluation_step_amp,
    supervised_training_step_tpu,
)
from ignite.metrics import MeanSquaredError
from tests.ignite import is_mps_available_and_functional


class DummyModel(torch.nn.Module):
    def __init__(self, output_as_list=False):
        super().__init__()
        self.output_as_list = output_as_list
        self.fc = torch.nn.Linear(1, 1, bias=False)

    def forward(self, x, bias=None):
        if bias is None:
            bias = 0.0
        if self.output_as_list:
            return self.fc(x) + bias, self.fc(x) + bias

        return self.fc(x) + bias


def _default_create_supervised_trainer(
    gradient_accumulation_steps: int = 1,
    model_device: str | None = None,
    trainer_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
    scaler: bool | GradScaler = False,
    with_model_transform: bool = False,
    with_model_fn: bool = False,
):
    if with_model_transform:

        def get_first_element(output):
            return output[0]

        model = DummyModel(output_as_list=True)
        model_transform = get_first_element
    else:
        model = DummyModel()
        model_transform = None

    if model_device:
        model.to(model_device)

    model.fc.weight.data.zero_()
    optimizer = SGD(model.parameters(), 0.1)

    if trace:
        example_inputs = (torch.randn(1), torch.randn(1)) if with_model_fn else torch.randn(1)
        model = torch.jit.trace(model, example_inputs)

    if amp_mode == "apex" and model_device == trainer_device == "cuda":
        from apex import amp

        model, optimizer = amp.initialize(model, optimizer, opt_level="O2")

    trainer = create_supervised_trainer(
        model,
        optimizer,
        mse_loss,
        device=trainer_device,
        output_transform=lambda x, y, y_pred, loss: (y_pred, loss.item()),
        amp_mode=amp_mode,
        scaler=scaler,
        gradient_accumulation_steps=gradient_accumulation_steps,
        model_transform=model_transform if model_transform is not None else lambda x: x,
        model_fn=(
            (lambda model, x: model(x, torch.tensor([0.01], device=model_device)))
            if with_model_fn
            else (lambda model, x: model(x))
        ),
    )
    assert model.fc.weight.data[0, 0].item() == approx(0.0)
    return trainer, model


def _test_create_supervised_trainer(
    gradient_accumulation_steps: int = 1,
    model_device: str | None = None,
    trainer_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
    scaler: bool | GradScaler = False,
    with_model_transform: bool = False,
    with_model_fn: bool = False,
):
    trainer, model = _default_create_supervised_trainer(
        gradient_accumulation_steps=gradient_accumulation_steps,
        model_device=model_device,
        trainer_device=trainer_device,
        trace=trace,
        amp_mode=amp_mode,
        scaler=scaler,
        with_model_transform=with_model_transform,
        with_model_fn=with_model_fn,
    )

    x = torch.tensor([[0.01], [0.02], [0.03], [0.04], [0.05]])
    y = torch.tensor([[0.015], [0.025], [0.035], [0.045], [0.055]])
    if with_model_fn:
        y += 0.01
    data = [(_x, _y) for _x, _y in zip(x, y)]

    theta = [0.0]
    accumulation = [0.0]
    loss = [0.0]

    @trainer.on(Events.ITERATION_COMPLETED)
    def _():
        assert model.fc.weight.grad != 0
        _x, _y = trainer.state.batch
        _x, _y = _x.to(model_device), _y.to(model_device)
        bias = 0.01 if with_model_fn else 0.0
        accumulation[0] += 0.2 * _x.item() * (theta[0] * _x.item() - (_y.item() - bias))
        # value of loss should not be accumulated
        _y_pred = model(_x, torch.tensor([bias], device=model_device)) if with_model_fn else model(_x)
        if with_model_transform:
            _y_pred = _y_pred[0]

        loss[0] = mse_loss(_y_pred, _y).item()

    @trainer.on(Events.ITERATION_COMPLETED(every=gradient_accumulation_steps))
    def _():
        theta[0] -= accumulation[0] / gradient_accumulation_steps
        assert pytest.approx(model.fc.weight.data[0, 0].item(), abs=1.0e-5) == theta[0]
        assert pytest.approx(trainer.state.output[-1], abs=1e-5) == loss[0]
        accumulation[0] = loss[0] = 0.0

    if model_device == trainer_device or ((model_device == "cpu") ^ (trainer_device == "cpu")):
        state = trainer.run(data)

        if amp_mode == "amp":
            assert state.output[0].dtype is torch.half
            if scaler and isinstance(scaler, bool):
                assert hasattr(state, "scaler")
            else:
                assert not hasattr(state, "scaler")

    else:
        if Version(torch.__version__) >= Version("1.7.0"):
            # This is broken in 1.6.0 but will be probably fixed with 1.7.0
            with pytest.raises(RuntimeError, match=r"Expected all tensors to be on the same device"):
                trainer.run(data)


@pytest.mark.skipif(Version(torch.__version__) < Version("2.3.1"), reason="Skip if < 2.3.1")
def test_create_supervised_training_scalar_assignment():
    with mock.patch("ignite.engine._check_arg") as check_arg_mock:
        check_arg_mock.return_value = None, GradScaler(enabled=False)
        trainer, _ = _default_create_supervised_trainer(model_device="cpu", trainer_device="cpu", scaler=True)
        assert hasattr(trainer.state, "scaler")
        assert isinstance(trainer.state.scaler, GradScaler)


def _test_create_mocked_supervised_trainer(
    model_device: str | None = None,
    trainer_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
    scaler: bool | GradScaler = False,
):
    with mock.patch("ignite.engine.supervised_training_step_amp") as training_step_amp_mock:
        with mock.patch("ignite.engine.supervised_training_step_apex") as training_step_apex_mock:
            with mock.patch("ignite.engine.supervised_training_step_tpu") as training_step_tpu_mock:
                with mock.patch("ignite.engine.supervised_training_step") as training_step_mock:
                    trainer, _ = _default_create_supervised_trainer(
                        model_device=model_device,
                        trainer_device=trainer_device,
                        trace=trace,
                        amp_mode=amp_mode,
                        scaler=scaler,
                    )

                    x = torch.tensor([[0.1], [0.2]])
                    y = torch.tensor([[0.3], [0.5]])
                    data = [(x, y)]

                    on_tpu = "xla" in trainer_device if trainer_device is not None else False
                    on_mps = "mps" in trainer_device if trainer_device is not None else False
                    mode, _ = _check_arg(on_tpu, on_mps, amp_mode, scaler)

                    if model_device == trainer_device or ((model_device == "cpu") ^ (trainer_device == "cpu")):
                        trainer.run(data)

                        if mode == "amp":
                            assert training_step_amp_mock.called
                        elif mode == "apex":
                            assert training_step_apex_mock.called
                        elif mode == "tpu":
                            assert training_step_tpu_mock.called
                        else:
                            assert training_step_mock.called


def _test_create_supervised_trainer_wrong_accumulation(
    model_device=None, trainer_device=None, amp_mode=None, trace=False
):
    with pytest.raises(ValueError, match="Gradient_accumulation_steps must be strictly positive."):
        _default_create_supervised_trainer(
            gradient_accumulation_steps=0,
            model_device=model_device,
            trainer_device=trainer_device,
            amp_mode=amp_mode,
            trace=trace,
        )


def _default_create_supervised_evaluator(
    model_device: str | None = None,
    evaluator_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
    with_model_transform: bool = False,
    with_model_fn: bool = False,
):
    if with_model_transform:

        def get_first_element(output):
            return output[0]

        model = DummyModel(output_as_list=True)
        model_transform = get_first_element
    else:
        model = DummyModel()
        model_transform = None

    if model_device:
        model.to(model_device)

    model.fc.weight.data.zero_()

    if trace:
        example_inputs = (torch.randn(1), torch.randn(1)) if with_model_fn else torch.randn(1)
        model = torch.jit.trace(model, example_inputs)

    evaluator = create_supervised_evaluator(
        model,
        device=evaluator_device,
        amp_mode=amp_mode,
        model_transform=model_transform if model_transform is not None else lambda x: x,
        model_fn=(
            (lambda model, x: model(x, torch.tensor([0.01], device=model_device)))
            if with_model_fn
            else (lambda model, x: model(x))
        ),
    )

    assert model.fc.weight.data[0, 0].item() == approx(0.0)

    return model, evaluator


def _test_create_supervised_evaluator(
    model_device: str | None = None,
    evaluator_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
    with_model_transform: bool = False,
    with_model_fn: bool = False,
):
    model, evaluator = _default_create_supervised_evaluator(
        model_device=model_device,
        evaluator_device=evaluator_device,
        trace=trace,
        amp_mode=amp_mode,
        with_model_transform=with_model_transform,
        with_model_fn=with_model_fn,
    )
    x = torch.tensor([[1.0], [2.0]])
    y = torch.tensor([[3.0], [5.0]])
    if with_model_fn:
        y += 0.01
    data = [(x, y)]

    if model_device == evaluator_device or ((model_device == "cpu") ^ (evaluator_device == "cpu")):
        state = evaluator.run(data)

        y_pred, y = state.output
        if with_model_fn:
            y_pred -= 0.01
            y -= 0.01
        assert y_pred[0, 0].item() == approx(0.0)
        assert y_pred[1, 0].item() == approx(0.0)
        assert y[0, 0].item() == approx(3.0)
        assert y[1, 0].item() == approx(5.0)

        assert model.fc.weight.data[0, 0].item() == approx(0.0)

    else:
        if Version(torch.__version__) >= Version("1.7.0"):
            # This is broken in 1.6.0 but will be probably fixed with 1.7.0
            err_msg_1 = "Expected all tensors to be on the same device"
            err_msg_2 = "Placeholder storage has not been allocated on MPS device"
            err_msg_3 = "Tensor for argument weight is on cpu but expected on mps"
            with pytest.raises(RuntimeError, match=f"({err_msg_1}|{err_msg_2}|{err_msg_3})"):
                evaluator.run(data)


def _test_mocked_supervised_evaluator(
    model_device: str | None = None,
    evaluator_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
):
    with mock.patch("ignite.engine.supervised_evaluation_step") as evaluation_step:
        with mock.patch("ignite.engine.supervised_evaluation_step_amp") as evaluation_step_amp:
            _, evaluator = _default_create_supervised_evaluator(
                model_device=model_device, evaluator_device=evaluator_device, trace=trace, amp_mode=amp_mode
            )

            x = torch.tensor([[1.0], [2.0]])
            y = torch.tensor([[3.0], [5.0]])
            data = [(x, y)]

            if model_device == evaluator_device or ((model_device == "cpu") ^ (evaluator_device == "cpu")):
                evaluator.run(data)

                if amp_mode == "amp":
                    assert evaluation_step_amp.called
                    assert not evaluation_step.called
                else:
                    assert evaluation_step.called
                    assert not evaluation_step_amp.called


def _test_create_evaluation_step_amp(
    autocast_mock,
    model_device: str | None = None,
    evaluator_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
):
    output_transform_mock = MagicMock()
    model = DummyModel()

    if model_device:
        model.to(model_device)

    model.fc.weight.data.zero_()

    if trace:
        example_input = torch.randn(1, 1)
        model = torch.jit.trace(model, example_input)

    device_type = evaluator_device.type if isinstance(evaluator_device, torch.device) else evaluator_device
    on_tpu = "xla" in device_type if device_type is not None else False
    on_mps = "mps" in device_type if device_type is not None else False
    mode, _ = _check_arg(on_tpu, on_mps, amp_mode, None)

    evaluate_step = supervised_evaluation_step_amp(model, evaluator_device, output_transform=output_transform_mock)

    x = torch.tensor([[1.0], [2.0]])
    y = torch.tensor([[3.0], [5.0]])
    data = [(x, y)]
    evaluator = Engine(evaluate_step)

    evaluator.run(data)
    assert autocast_mock.called
    assert output_transform_mock.called


def _test_create_evaluation_step(
    mock_torch_cuda_amp_module,
    model_device: str | None = None,
    evaluator_device: str | None = None,
    trace: bool = False,
    amp_mode: str = None,
):
    output_transform_mock = MagicMock()
    model = DummyModel()

    if model_device:
        model.to(model_device)

    model.fc.weight.data.zero_()

    if trace:
        example_input = torch.randn(1, 1)
        model = torch.jit.trace(model, example_input)

    device_type = evaluator_device.type if isinstance(evaluator_device, torch.device) else evaluator_device
    on_tpu = "xla" in device_type if device_type is not None else False
    on_mps = "mps" in device_type if device_type is not None else False
    mode, _ = _check_arg(on_tpu, on_mps, amp_mode, None)

    evaluate_step = supervised_evaluation_step(model, evaluator_device, output_transform=output_transform_mock)

    x = torch.tensor([[1.0], [2.0]])
    y = torch.tensor([[3.0], [5.0]])
    data = [(x, y)]
    evaluator = Engine(evaluate_step)

    evaluator.run(data)
    assert not mock_torch_cuda_amp_module.called
    assert output_transform_mock.called


@pytest.mark.parametrize("trainer_device", [None, "cpu"])
@pytest.mark.parametrize("trace", [False, True])
def test_create_supervised_trainer(trainer_device, trace):
    _test_create_supervised_trainer_wrong_accumulation(trainer_device=trainer_device, trace=trace)
    _test_create_supervised_trainer(gradient_accumulation_steps=1, trainer_device=trainer_device, trace=trace)
    _test_create_supervised_trainer(gradient_accumulation_steps=3, trainer_device=trainer_device, trace=trace)
    _test_create_supervised_trainer(with_model_transform=True, trainer_device=trainer_device, trace=trace)
    _test_create_supervised_trainer(with_model_fn=True, trainer_device=trainer_device, trace=trace)
    _test_create_mocked_supervised_trainer(trainer_device=trainer_device, trace=trace)


@pytest.mark.skipif(find_spec("apex"), reason="Skip if APEX")
def test_create_supervised_trainer_apex_error():
    with pytest.raises(
        ModuleNotFoundError, match="Please install apex from https://github.com/nvidia/apex to use amp_mode='apex'."
    ):
        _test_create_supervised_trainer_wrong_accumulation(trainer_device="cpu", amp_mode="apex")
    with pytest.raises(
        ModuleNotFoundError, match="Please install apex from https://github.com/nvidia/apex to use amp_mode='apex'."
    ):
        _test_create_supervised_trainer(amp_mode="apex")


@pytest.fixture
def mock_torch_cuda_amp_module():
    with patch.dict(
        "sys.modules",
        {"torch.amp": None, "torch.cuda.amp": None, "torch.amp.autocast_mode": None},
    ):
        yield torch


def test_create_supervised_trainer_amp_error(mock_torch_cuda_amp_module):
    with pytest.raises(ImportError, match="Please install torch>=1.12.0 to use amp_mode='amp'."):
        _test_create_supervised_trainer_wrong_accumulation(trainer_device="cpu", amp_mode="amp")
    with pytest.raises(ImportError, match="Please install torch>=1.12.0 to use amp_mode='amp'."):
        _test_create_supervised_trainer(amp_mode="amp")
    with pytest.raises(ImportError, match="Please install torch>=2.3.1 to use scaler argument."):
        _test_create_supervised_trainer(amp_mode="amp", scaler=True)


@pytest.mark.skipif(Version(torch.__version__) < Version("2.3.1"), reason="Skip if < 2.3.1")
def test_create_supervised_trainer_scaler_not_amp():
    scaler = GradScaler(enabled=torch.cuda.is_available())

    with pytest.raises(ValueError, match=f"scaler argument is {scaler}, but amp_mode is None."):
        _test_create_supervised_trainer(amp_mode=None, scaler=scaler)
    with pytest.raises(ValueError, match="scaler argument is True, but amp_mode is None."):
        _test_create_supervised_trainer(amp_mode=None, scaler=True)
    with pytest.raises(ValueError, match="scaler argument is True, but amp_mode is apex."):
        _test_create_supervised_trainer(amp_mode="apex", scaler=True)
    with pytest.raises(ValueError, match=f"scaler argument is {scaler}, but amp_mode is apex."):
        _test_create_supervised_trainer(amp_mode="apex", scaler=scaler)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_trainer_on_cuda():
    model_device = trainer_device = "cuda"
    _test_create_supervised_trainer_wrong_accumulation(model_device=model_device, trainer_device=trainer_device)
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1, model_device=model_device, trainer_device=trainer_device
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3, model_device=model_device, trainer_device=trainer_device
    )
    _test_create_mocked_supervised_trainer(model_device=model_device, trainer_device=trainer_device)


@pytest.mark.skipif(not is_mps_available_and_functional(), reason="Skip if no MPS")
def test_create_supervised_trainer_on_mps():
    model_device = trainer_device = "mps"
    _test_create_supervised_trainer_wrong_accumulation(model_device=model_device, trainer_device=trainer_device)
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1, model_device=model_device, trainer_device=trainer_device
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3, model_device=model_device, trainer_device=trainer_device
    )
    _test_create_mocked_supervised_trainer(model_device=model_device, trainer_device=trainer_device)


@pytest.mark.skipif(Version(torch.__version__) < Version("1.12.0"), reason="Skip if < 1.12.0")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_trainer_on_cuda_amp():
    model_device = trainer_device = "cuda"
    _test_create_supervised_trainer_wrong_accumulation(
        model_device=model_device, trainer_device=trainer_device, amp_mode="amp"
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1, model_device=model_device, trainer_device=trainer_device, amp_mode="amp"
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3, model_device=model_device, trainer_device=trainer_device, amp_mode="amp"
    )
    _test_create_mocked_supervised_trainer(model_device=model_device, trainer_device=trainer_device, amp_mode="amp")


@pytest.mark.skipif(Version(torch.__version__) < Version("1.12.0"), reason="Skip if < 1.12.0")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_trainer_on_cuda_amp_scaler():
    model_device = trainer_device = "cuda"
    _test_create_supervised_trainer_wrong_accumulation(
        model_device=model_device, trainer_device=trainer_device, amp_mode="amp"
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1,
        model_device=model_device,
        trainer_device=trainer_device,
        amp_mode="amp",
        scaler=True,
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3,
        model_device=model_device,
        trainer_device=trainer_device,
        amp_mode="amp",
        scaler=True,
    )
    _test_create_mocked_supervised_trainer(
        model_device=model_device, trainer_device=trainer_device, amp_mode="amp", scaler=True
    )
    scaler = GradScaler(enabled=torch.cuda.is_available())
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1,
        model_device=model_device,
        trainer_device=trainer_device,
        amp_mode="amp",
        scaler=scaler,
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3,
        model_device=model_device,
        trainer_device=trainer_device,
        amp_mode="amp",
        scaler=scaler,
    )
    _test_create_mocked_supervised_trainer(
        model_device=model_device, trainer_device=trainer_device, amp_mode="amp", scaler=scaler
    )


# @pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
# @pytest.mark.skipif(not find_spec("apex"), reason="Skip if no APEX")
@pytest.mark.skip(reason="Temporarily disabled, as it fails because of an issue from apex side")
def test_create_supervised_trainer_on_cuda_apex():
    model_device = trainer_device = "cuda"
    _test_create_supervised_trainer_wrong_accumulation(
        model_device=model_device, trainer_device=trainer_device, amp_mode="apex"
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1, model_device=model_device, trainer_device=trainer_device, amp_mode="apex"
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3, model_device=model_device, trainer_device=trainer_device, amp_mode="apex"
    )
    _test_create_mocked_supervised_trainer(model_device=model_device, trainer_device=trainer_device, amp_mode="apex")


@pytest.mark.skipif(idist.has_xla_support, reason="Skip if has PyTorch XLA package")
def test_supervised_training_step_tpu_no_xla():
    with pytest.raises(ModuleNotFoundError, match="torch_xla cannot be imported, please install PyTorch XLA."):
        supervised_training_step_tpu(model=None, optimizer=None, loss_fn=None)


@pytest.mark.skipif(idist.has_xla_support, reason="Skip if has PyTorch XLA package")
def test_create_supervised_trainer_on_tpu_no_xla():
    model_device = "cpu"
    trainer_device = "xla"
    with pytest.raises(RuntimeError, match=r"In order to run on TPU, please install PyTorch XLA"):
        _test_create_supervised_trainer(model_device=model_device, trainer_device=trainer_device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_create_supervised_trainer_on_tpu():
    model_device = trainer_device = "xla"
    _test_create_supervised_trainer_wrong_accumulation(model_device=model_device, trainer_device=trainer_device)
    _test_create_supervised_trainer(
        gradient_accumulation_steps=1, model_device=model_device, trainer_device=trainer_device
    )
    _test_create_supervised_trainer(
        gradient_accumulation_steps=3, model_device=model_device, trainer_device=trainer_device
    )
    _test_create_mocked_supervised_trainer(model_device=model_device, trainer_device=trainer_device)


@pytest.mark.tpu
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_create_supervised_trainer_on_tpu_amp():
    model_device = trainer_device = "xla"
    with pytest.raises(ValueError, match="amp_mode cannot be used with xla device."):
        _test_create_supervised_trainer(model_device=model_device, trainer_device=trainer_device, amp_mode="amp")


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_trainer_on_cuda_with_model_on_cpu():
    _test_create_supervised_trainer_wrong_accumulation(trainer_device="cuda")
    _test_create_supervised_trainer(gradient_accumulation_steps=1, trainer_device="cuda")
    _test_create_supervised_trainer(gradient_accumulation_steps=3, trainer_device="cuda")
    _test_create_mocked_supervised_trainer(trainer_device="cuda")


def test_create_supervised_evaluator():
    _test_create_supervised_evaluator()
    _test_create_supervised_evaluator(with_model_transform=True)
    _test_create_supervised_evaluator(with_model_fn=True)
    _test_mocked_supervised_evaluator()

    # older versions didn't have the autocast method so we skip the test for older builds
    if Version(torch.__version__) >= Version("1.12.0"):
        with mock.patch("torch.amp.autocast") as mock_torch_cuda_amp_module:
            _test_create_evaluation_step_amp(mock_torch_cuda_amp_module)


def test_create_supervised_evaluator_on_cpu():
    _test_create_supervised_evaluator(evaluator_device="cpu")
    _test_mocked_supervised_evaluator(evaluator_device="cpu")

    # older versions didn't have the autocast method so we skip the test for older builds
    if Version(torch.__version__) >= Version("1.12.0"):
        with mock.patch("torch.amp.autocast") as mock_torch_cuda_amp_module:
            _test_create_evaluation_step(mock_torch_cuda_amp_module, evaluator_device="cpu")
            _test_create_evaluation_step_amp(mock_torch_cuda_amp_module, evaluator_device="cpu")


def test_create_supervised_evaluator_traced_on_cpu():
    _test_create_supervised_evaluator(evaluator_device="cpu", trace=True)
    _test_mocked_supervised_evaluator(evaluator_device="cpu", trace=True)

    # older versions didn't have the autocast method so we skip the test for older builds
    if Version(torch.__version__) >= Version("1.12.0"):
        with mock.patch("torch.amp.autocast") as mock_torch_cuda_amp_module:
            _test_create_evaluation_step(mock_torch_cuda_amp_module, evaluator_device="cpu", trace=True)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_evaluator_on_cuda():
    model_device = evaluator_device = "cuda"
    _test_create_supervised_evaluator(model_device=model_device, evaluator_device=evaluator_device)
    _test_mocked_supervised_evaluator(model_device=model_device, evaluator_device=evaluator_device)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_evaluator_on_cuda_with_model_on_cpu():
    _test_create_supervised_evaluator(evaluator_device="cuda")
    _test_mocked_supervised_evaluator(evaluator_device="cuda")


@pytest.mark.skipif(not is_mps_available_and_functional(), reason="Skip if no MPS")
def test_create_supervised_evaluator_on_mps():
    model_device = evaluator_device = "mps"
    _test_create_supervised_evaluator(model_device=model_device, evaluator_device=evaluator_device)
    _test_mocked_supervised_evaluator(model_device=model_device, evaluator_device=evaluator_device)


@pytest.mark.skipif(not is_mps_available_and_functional(), reason="Skip if no MPS")
def test_create_supervised_evaluator_on_mps_with_model_on_cpu():
    _test_create_supervised_evaluator(evaluator_device="mps")
    _test_mocked_supervised_evaluator(evaluator_device="mps")


@pytest.mark.skipif(Version(torch.__version__) < Version("1.12.0"), reason="Skip if < 1.12.0")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_evaluator_on_cuda_amp():
    model_device = evaluator_device = "cuda"
    _test_create_supervised_evaluator(model_device=model_device, evaluator_device=evaluator_device, amp_mode="amp")
    _test_mocked_supervised_evaluator(model_device=model_device, evaluator_device=evaluator_device, amp_mode="amp")


def test_create_supervised_evaluator_amp_error(mock_torch_cuda_amp_module):
    with pytest.raises(ImportError, match="Please install torch>=1.12.0 to use amp_mode='amp'."):
        _test_create_supervised_evaluator(amp_mode="amp")


def test_create_supervised_evaluator_with_metrics():
    model = DummyModel()
    model.fc.weight.data.zero_()

    evaluator = create_supervised_evaluator(model, metrics={"mse": MeanSquaredError()})

    x = torch.tensor([[1.0], [2.0]])
    y = torch.tensor([[3.0], [4.0]])
    data = [(x, y)]

    state = evaluator.run(data)
    assert state.metrics["mse"] == 12.5



import pytest
import torch
from torch.testing import assert_close

import ignite.distributed as idist
from fairlearn.metrics import MetricFrame
from sklearn.metrics import accuracy_score
from ignite.exceptions import NotComputableError
from ignite.metrics.fairness.accuracy_difference import SubgroupAccuracyDifference


def test_subgroup_accuracy_difference_empty() -> None:
    """Tests if NotComputableError is raised when no data is provided."""
    metric = SubgroupAccuracyDifference(groups=[0, 1])
    with pytest.raises(NotComputableError, match="Fairness metrics must have at least one example"):
        metric.compute()


def test_subgroup_accuracy_difference_single_group() -> None:
    """Tests if NotComputableError is raised when only one subgroup is present."""
    metric = SubgroupAccuracyDifference(groups=[0])

    y_pred = torch.tensor([[0.9, 0.1], [0.8, 0.2]])
    y = torch.tensor([0, 0])
    group_labels = torch.tensor([0, 0])

    metric.update((y_pred, y, group_labels))

    with pytest.raises(NotComputableError, match="Fairness metrics require at least two unique subgroups"):
        metric.compute()


def test_subgroup_accuracy_difference_binary_labels(available_device) -> None:
    """Tests SubgroupAccuracyDifference with binary 0/1 labels."""
    metric = SubgroupAccuracyDifference(groups=[0, 1], device=available_device)

    # y_pred and y are (B,)
    # Group 0: 2/2 correct (1.0)
    # Group 1: 0/2 correct (0.0)
    y_pred = torch.tensor([1, 0, 1, 0], device=available_device)
    y = torch.tensor([1, 0, 0, 1], device=available_device)
    groups = torch.tensor([0, 0, 1, 1], device=available_device)

    metric.update((y_pred, y, groups))
    assert_close(metric.compute(), 1.0)


def test_subgroup_accuracy_difference_binary_probs(available_device) -> None:
    """Tests SubgroupAccuracyDifference with (B, 1) thresholded labels."""
    metric = SubgroupAccuracyDifference(groups=[0, 1], device=available_device)

    # y_pred is (B, 1), y is (B,)
    # Group 0: 2/2 correct (1.0)
    # Group 1: 1/2 correct (0.5)
    y_pred = torch.tensor([[1], [0], [1], [1]], device=available_device)
    y = torch.tensor([1, 0, 1, 0], device=available_device)
    groups = torch.tensor([0, 0, 1, 1], device=available_device)

    metric.update((y_pred, y, groups))
    assert_close(metric.compute(), 0.5)


def test_subgroup_accuracy_difference_multiclass(available_device) -> None:
    """Tests SubgroupAccuracyDifference with multiclass logits."""
    metric = SubgroupAccuracyDifference(groups=[0, 1], device=available_device)

    # y_pred is (B, C), y is (B,)
    # Group 0: 1/1 correct (1.0)
    # Group 1: 0/1 correct (0.0)
    y_pred = torch.tensor([[0.1, 0.8, 0.1], [0.8, 0.1, 0.1]], device=available_device)
    y = torch.tensor([1, 1], device=available_device)
    groups = torch.tensor([0, 1], device=available_device)

    metric.update((y_pred, y, groups))
    assert_close(metric.compute(), 1.0)


def test_subgroup_accuracy_difference_multilabel(available_device) -> None:
    """Tests SubgroupAccuracyDifference with multilabel data."""
    metric = SubgroupAccuracyDifference(groups=[0, 1], is_multilabel=True, device=available_device)

    # y_pred and y are (B, C)
    # Accuracy uses sample-wise correctness: all labels must match per sample.
    # Group 0 (sample 0): y_pred=[1,0], y=[1,0] -> all match -> correct. Accuracy = 1/1 = 1.0
    # Group 1 (sample 1): y_pred=[1,1], y=[1,0] -> not all match -> incorrect. Accuracy = 0/1 = 0.0
    y_pred = torch.tensor([[1, 0], [1, 1]], device=available_device)
    y = torch.tensor([[1, 0], [1, 0]], device=available_device)
    groups = torch.tensor([0, 1], device=available_device)

    metric.update((y_pred, y, groups))
    assert_close(metric.compute(), 1.0)


def test_subgroup_accuracy_difference_spatial(available_device) -> None:
    """Tests SubgroupAccuracyDifference with spatial (image) data."""
    metric = SubgroupAccuracyDifference(groups=[0, 1], device=available_device)

    # y_pred is (B, C, H, W), y is (B, H, W)
    # B=2, C=2, H=2, W=2
    # Group 0: spatial targets (4 pixels) all correct (1.0)
    # Group 1: spatial targets (4 pixels) all incorrect (0.0)

    y_pred = torch.zeros(2, 2, 2, 2, device=available_device)
    # Group 0 (index 0): predict class 0 for all
    y_pred[0, 0, :, :] = 1.0
    # Group 1 (index 1): predict class 0 for all
    y_pred[1, 0, :, :] = 1.0

    y = torch.zeros(2, 2, 2, dtype=torch.long, device=available_device)
    # Group 0: all class 0 (1.0 acc)
    y[0, :, :] = 0
    # Group 1: all class 1 (0.0 acc)
    y[1, :, :] = 1

    groups = torch.tensor([0, 1], device=available_device)

    metric.update((y_pred, y, groups))
    assert_close(metric.compute(), 1.0)


def test_compare_accuracy_difference_with_fairlearn(available_device) -> None:
    """Verifies SubgroupAccuracyDifference matches Fairlearn's MetricFrame(accuracy_score).difference()"""
    groups_list = [0, 1]
    ignite_metric = SubgroupAccuracyDifference(groups=groups_list, device=available_device)

    # Random binary data
    y_pred_probs = torch.tensor(
        [[0.1, 0.9], [0.8, 0.2], [0.3, 0.7], [0.6, 0.4], [0.2, 0.8], [0.7, 0.3]], device=available_device
    )
    y_pred = torch.argmax(y_pred_probs, dim=1)
    y_true = torch.tensor([1, 0, 1, 1, 0, 0], device=available_device)
    group_labels = torch.tensor([0, 0, 0, 1, 1, 1], device=available_device)

    # Ignite update and compute
    ignite_metric.update((y_pred_probs, y_true, group_labels))
    ignite_result = ignite_metric.compute()

    # Fairlearn computation
    # Fairlearn's MetricFrame takes numpy arrays
    mf = MetricFrame(
        metrics=accuracy_score,
        y_true=y_true.cpu().numpy(),
        y_pred=y_pred.cpu().numpy(),
        sensitive_features=group_labels.cpu().numpy(),
    )
    fairlearn_result = mf.difference()

    assert_close(ignite_result, float(fairlearn_result))


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_compute_subgroup_accuracy_difference(self) -> None:
        rank = idist.get_rank()
        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            metric = SubgroupAccuracyDifference(groups=[0, 1], device=metric_device)

            y = torch.tensor([0, 1], device=device)
            groups = torch.tensor([0, 1], device=device)

            if rank == 0:
                # Group 0: correct, Group 1: incorrect
                y_pred = torch.tensor([[0.8, 0.2], [0.8, 0.2]], device=device)
            else:
                # Group 0: correct, Group 1: incorrect
                y_pred = torch.tensor([[0.9, 0.1], [0.9, 0.1]], device=device)

            metric.update((y_pred, y, groups))
            res = metric.compute()

            # Across all ranks:
            # Group 0: all correct (1.0)
            # Group 1: all incorrect (0.0)
            assert_close(res, 1.0)


def test_subgroup_accuracy_difference_validation() -> None:
    """Tests input validation for SubgroupAccuracyDifference."""
    metric = SubgroupAccuracyDifference(groups=[0, 1])

    # Valid multiclass logits
    y_pred = torch.randn(4, 3)
    y = torch.randint(0, 3, (4,))
    groups = torch.tensor([0, 0, 1, 1])
    metric.update((y_pred, y, groups))
    assert metric.compute() >= 0

    metric.reset()

    # Valid binary classes (0/1)
    y_pred = torch.tensor([0, 1, 0, 1])
    y = torch.tensor([0, 1, 1, 0])
    groups = torch.tensor([0, 0, 1, 1])
    metric.update((y_pred, y, groups))
    assert metric.compute() == 1.0  # Subgroup 0: 100%, Subgroup 1: 0%

    metric.reset()

    # Invalid binary probabilities (not 0 or 1)
    y_pred = torch.tensor([0.6, 0.4, 0.8, 0.1])
    y = torch.tensor([1, 0, 1, 0])
    groups = torch.tensor([0, 0, 1, 1])
    with pytest.raises(ValueError, match="y_pred must be comprised of 0's and 1's"):
        metric.update((y_pred, y, groups))


def test_multilabel_validation() -> None:
    """Tests input validation for multilabel data."""
    metric = SubgroupAccuracyDifference(groups=[0, 1], is_multilabel=True)

    # Valid multilabel (0/1)
    y_pred = torch.tensor([[1, 0], [1, 1], [0, 0], [0, 1]])
    y = torch.tensor([[1, 0], [1, 1], [1, 1], [0, 0]])
    groups = torch.tensor([0, 0, 1, 1])
    metric.update((y_pred, y, groups))
    # Accuracy uses sample-wise correctness: all labels must match per sample.
    # Group 0 (samples 0,1): [1,0]==[1,0], [1,1]==[1,1] -> Accuracy = 2/2 = 1.0
    # Group 1 (samples 2,3): [0,0]!=[1,1], [0,1]!=[0,0] -> Accuracy = 0/2 = 0.0
    # Disparity = 1.0 - 0.0 = 1.0
    assert metric.compute() == 1.0

    metric.reset()

    # Invalid multilabel (not 0/1)
    y_pred = torch.tensor([[0.6, 0.4], [0.8, 0.1]])
    y = torch.tensor([[1, 0], [1, 1]])
    groups = torch.tensor([0, 1])
    with pytest.raises(ValueError, match="y_pred must be comprised of 0's and 1's"):
        metric.update((y_pred, y, groups))


def test_shape_mismatch_validation() -> None:
    """Tests validation for shape mismatches between y and y_pred."""
    metric = SubgroupAccuracyDifference(groups=[0, 1])

    # y is (B,), y_pred is (B, C) - OK
    metric.update((torch.randn(4, 3), torch.randint(0, 3, (4,)), torch.zeros(4)))

    # y is (B,), y_pred is (B, C, H, W) - Error (dimension mismatch)
    with pytest.raises(ValueError, match="y must have shape of"):
        metric.update((torch.randn(4, 3, 2, 2), torch.randint(0, 3, (4,)), torch.zeros(4)))


def test_batch_size_mismatch_validation() -> None:
    """Tests validation for batch size mismatches across y_pred, y, and groups."""
    metric = SubgroupAccuracyDifference(groups=[0, 1])

    # y_pred (4), y (3) - Batch size mismatch
    with pytest.raises(ValueError, match="y_pred, y, and group_labels must have the same batch size"):
        metric.update((torch.randn(4, 3), torch.randint(0, 3, (3,)), torch.zeros(4)))

    # groups (3), y_pred (4) - Batch size mismatch
    with pytest.raises(ValueError, match="y_pred, y, and group_labels must have the same batch size"):
        metric.update((torch.randn(4, 3), torch.randint(0, 3, (4,)), torch.zeros(3)))


def test_type_switch_validation() -> None:
    """Tests if RuntimeError is raised when input type changes mid-epoch."""
    metric = SubgroupAccuracyDifference(groups=[0, 1])

    # First batch: binary
    metric.update((torch.tensor([1, 0]), torch.tensor([1, 0]), torch.tensor([0, 1])))

    # Second batch: multiclass - Should raise RuntimeError
    with pytest.raises(RuntimeError, match="Input data type has changed from binary to multiclass"):
        metric.update((torch.randn(2, 3), torch.tensor([0, 1]), torch.tensor([0, 1])))

    # Third batch: multiclass (3) then multiclass (5) - Should raise ValueError
    metric.reset()
    metric.update((torch.randn(4, 3), torch.randint(0, 3, (4,)), torch.zeros(4)))
    with pytest.raises(ValueError, match="Input data number of classes has changed"):
        metric.update((torch.randn(4, 5), torch.randint(0, 5, (4,)), torch.zeros(4)))



import argparse
from pathlib import Path

import torch
import torch.nn as nn
import torch.optim as optim

import ray
from ray import tune
from ray.tune import Checkpoint as RayCheckpoint, CLIReporter
from ray.tune.schedulers import ASHAScheduler

from ignite.engine import Events, create_supervised_trainer, create_supervised_evaluator
from ignite.metrics import Loss, Accuracy
from ignite.handlers import Checkpoint, DiskSaver

from utils import Net, get_data_loaders, get_test_loader, load_data


def train_with_ignite(config, data_dir=None, num_epochs=10, num_workers=8):
    device = config.get("device", "cpu")
    net = Net(config["l1"], config["l2"])
    net = net.to(device)

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(net.parameters(), lr=config["lr"], momentum=0.9)

    train_loader, val_loader = get_data_loaders(config["batch_size"], data_dir, num_workers)

    trainer = create_supervised_trainer(net, optimizer, criterion, device=device)

    metrics = {"accuracy": Accuracy(), "loss": Loss(criterion)}
    evaluator = create_supervised_evaluator(net, metrics=metrics, device=device)

    checkpoint_dir = tune.get_context().get_trial_dir()
    save_handler = DiskSaver(checkpoint_dir, create_dir=True)
    to_save = {"net": net, "optimizer": optimizer, "trainer": trainer}
    checkpoint_handler = Checkpoint(to_save, save_handler=save_handler, filename_pattern="checkpoint.pt")
    trainer.add_event_handler(Events.EPOCH_COMPLETED, checkpoint_handler)

    @trainer.on(Events.EPOCH_COMPLETED)
    def log_and_report(engine):
        epoch = engine.state.epoch

        evaluator.run(val_loader)
        val_loss = evaluator.state.metrics["loss"]
        val_accuracy = evaluator.state.metrics["accuracy"]

        print(f"Epoch {epoch}: loss={val_loss:.4f}, accuracy={val_accuracy:.4f}")

        checkpoint = RayCheckpoint.from_directory(checkpoint_dir)
        tune.report(
            {"loss": val_loss, "accuracy": val_accuracy},
            checkpoint=checkpoint,
        )

    trainer.run(train_loader, num_epochs)


def test_accuracy(model, device, data_dir):
    test_loader = get_test_loader(4, data_dir)
    model = model.to(device)

    evaluator = create_supervised_evaluator(model, metrics={"accuracy": Accuracy()}, device=device)
    evaluator.run(test_loader)

    return evaluator.state.metrics["accuracy"]


def tune_cifar(num_trials=10, num_epochs=10, gpus_per_trial=1, cpus_per_trial=4, data_dir="./data"):
    data_dir = Path(data_dir).resolve()

    config = {
        "l1": tune.choice([2**i for i in range(2, 9)]),
        "l2": tune.choice([2**i for i in range(2, 9)]),
        "lr": tune.loguniform(1e-4, 1e-1),
        "batch_size": tune.choice([256, 512, 1024]),
        "device": "cuda" if gpus_per_trial > 0 else "cpu",
    }

    scheduler = ASHAScheduler(
        max_t=num_epochs,
        grace_period=1,
        reduction_factor=2,
    )

    print("Downloading CIFAR10 dataset...")

    load_data(data_dir)

    reporter = CLIReporter(
        metric_columns={"loss": "loss", "accuracy": "accuracy"},
        parameter_columns={"l1": "l1", "l2": "l2", "lr": "lr", "batch_size": "batch_size"},
    )

    tuner = tune.Tuner(
        tune.with_resources(
            tune.with_parameters(
                train_with_ignite,
                data_dir=data_dir,
                num_epochs=num_epochs,
            ),
            resources={"cpu": cpus_per_trial, "gpu": gpus_per_trial},
        ),
        tune_config=tune.TuneConfig(
            metric="loss",
            mode="min",
            scheduler=scheduler,
            num_samples=num_trials,
        ),
        param_space=config,
        run_config=ray.tune.RunConfig(
            storage_path=None,
            name="cifar10_ray_tune",
            progress_reporter=reporter,
        ),
    )

    results = tuner.fit()

    best_result = results.get_best_result("accuracy", "max", filter_nan_and_inf=False)
    print(f"Best trial config: {best_result.config}")
    print(f"Best trial final validation loss: {best_result.metrics['loss']}")
    print(f"Best trial final validation accuracy: {best_result.metrics['accuracy']}")

    best_trained_model = Net(best_result.config["l1"], best_result.config["l2"])
    device = best_result.config["device"]
    best_trained_model = best_trained_model.to(device)

    best_checkpoint = best_result.checkpoint
    with best_checkpoint.as_directory() as checkpoint_dir:
        checkpoint_path = Path(checkpoint_dir) / "checkpoint.pt"
        best_checkpoint_data = torch.load(checkpoint_path, map_location=device)
        best_trained_model.load_state_dict(best_checkpoint_data["net"])

        test_acc = test_accuracy(best_trained_model, device, data_dir)
        print(f"Best trial test set accuracy: {test_acc}")

    return results


def main():
    parser = argparse.ArgumentParser(description="Ignite + Ray Tune CIFAR10 Example")
    parser.add_argument("--num_trials", type=int, default=10, help="Number of trials for Ray Tune")
    parser.add_argument("--num_epochs", type=int, default=10, help="Number of epochs per trial")
    parser.add_argument("--gpus_per_trial", type=float, default=1, help="GPUs per trial")
    parser.add_argument("--cpus_per_trial", type=int, default=4, help="CPUs per trial")
    parser.add_argument(
        "--data_dir",
        type=str,
        default="./data",
        help="Path to CIFAR10 dataset",
    )
    args = parser.parse_args()

    ray.init(include_dashboard=False)
    tune_cifar(
        num_trials=args.num_trials,
        num_epochs=args.num_epochs,
        gpus_per_trial=args.gpus_per_trial,
        cpus_per_trial=args.cpus_per_trial,
        data_dir=args.data_dir,
    )


if __name__ == "__main__":
    main()
